from django.conf import settings
from django.conf.urls import url
from django.conf.urls.static import static

from . import views

urlpatterns = static(
    settings.MEDIA_URL, document_root=settings.MEDIA_ROOT) + [
        url(
            r'^archive/(?P<identifier>[\w\-]{20,35})/$',
            views.hosted_message, name='message_web_view')]



from django.contrib import admin
from django.conf.urls import url
from django.conf.urls import include
from django.conf.urls.i18n import i18n_patterns

import rest_framework.urls
from rest_framework_jwt.views import refresh_jwt_token

import munch.apps.abuse.urls
import munch.apps.users.urls
import munch.apps.hosted.urls
import munch.apps.optouts.urls
import munch.apps.tracking.urls

from munch.core.views import api_root
from munch.apps.users.api.v1.views import ObtainJSONWebToken

api_urlpatterns_v1 = [
    url(r'^$', api_root, name='api-root'),
]

api_urlpatterns_v2 = [
    url(r'^$', api_root, name='api-root'),
]

urlpatterns = [
    url(r'^auth/', include(
        rest_framework.urls, namespace='rest_framework')),
    url(r'^api-token-auth', ObtainJSONWebToken.as_view()),
    url(r'^api-token-refresh', refresh_jwt_token),
    url(r'^abuse/', include(munch.apps.abuse.urls)),
    url(r'^t/', include(munch.apps.tracking.urls)),
    url(r'^h/', include(munch.apps.optouts.urls)),
    url(r'^account/', include(munch.apps.users.urls)),
    url(r'', include(munch.apps.hosted.urls)),
    url(r'^v1/', include(api_urlpatterns_v1, namespace='v1')),
    # url(r'^v2/', include(api_urlpatterns_v2, namespace='v2'))
]

urlpatterns += i18n_patterns(url(r'^admin/', admin.site.urls))



import django_filters

from munch.core.utils.filters import DRFFilterSet

from ...models import Mail
from ...models import MailStatus


class MailFilter(DRFFilterSet):
    class Meta:
        model = Mail
        fields = ['batch']
        order_by = True


class MailStatusFilter(DRFFilterSet):
    address = django_filters.CharFilter(name='mail__recipient')

    class Meta:
        model = MailStatus
        fields = ['address', 'creation_date', 'status']
        order_by = True



from .base import *
from urllib.parse import urlparse

##########
# Django #
##########

INSTALLED_APPS += ['corsheaders']

MIDDLEWARE_CLASSES.insert(MIDDLEWARE_CLASSES.index('django.middleware.common.CommonMiddleware') - 1, 'corsheaders.middleware.CorsMiddleware')  # noqa
CORS_ORIGIN_ALLOW_ALL = True

DEBUG = True
SECRET_KEY = 'test'
# CONTAINER_IP = urlparse(os.environ.get(
#     'DOCKER_HOST', 'tcp://localhost')).hostname
CONTAINER_IP = '127.0.0.1'
BROKER_URL = 'amqp://guest:guest@{}:5682/munch'.format(CONTAINER_IP)
DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.postgresql_psycopg2',
        'NAME': 'munch',
        'USER': 'munch',
        'PASSWORD': 'munch',
        'HOST': CONTAINER_IP,
        'PORT': '15432',
    }
}

CACHES = {
    "default": {
        "BACKEND": "django_redis.cache.RedisCache",
        "LOCATION": "redis://{}:16379/1".format(CONTAINER_IP),
        "OPTIONS": {
            "CLIENT_CLASS": "django_redis.client.DefaultClient",
        }
    }
}

###########
# Logging #
###########
LOGGING['loggers']['munch'] = {
    'handlers': ['console'],
    'filters': [],
    'propagate': False,
    'level': logging.DEBUG,
}
LOGGING['loggers']['mailsend'] = {
    'handlers': ['console'],
    'level': logging.DEBUG,
    'propagate': False
}

##########
# Global #
##########
APPLICATION_URL = 'http://localhost:8000'
MASS_SENDER_DOMAIN = 'localhost'
MASS_EMAIL_BACKEND = 'munch.core.mail.backend.DummyBackend'
BYPASS_DNS_CHECKS = True

#################
# Transactional #
#################
TRANSACTIONAL['SMTP_BIND_HOST'] = '0.0.0.0'
TRANSACTIONAL['SMTP_BIND_PORT'] = 1025
TRANSACTIONAL['SMTP_REQUIRE_CREDENTIALS'] = True
TRANSACTIONAL['SMTP_CREDENTIALS'] = {'admin': 'admin'}
TRANSACTIONAL['SMTP_SMARTHOST_TLS'] = {
    'keyfile': '<path_to>/munch/utils/ssl/postfix.example.com.key.pem',
    'certfile': '<path_to>/munch/utils/ssl/postfix.example.com.cert.pem',
}

#############
# Campaigns #
#############
CAMPAIGNS['SKIP_SPAM_CHECK'] = True
CAMPAIGNS['SKIP_VIRUS_CHECK'] = True
CAMPAIGNS['BYPASS_RECIPIENTS_MX_CHECK'] = True

SPAMD_PORT = 1783
SPAMD_HOST = 'localhost'
CLAMD_PORT = 13310
CLAMD_HOST = 'localhost'



from rest_framework import serializers

from ...models import OptOut


class OptOutSerializer(serializers.ModelSerializer):
    class Meta:
        model = OptOut
        fields = ['identifier', 'address', 'creation_date', 'origin']
        read_only_fields = ['identifier', 'address']



# -*- coding: utf-8 -*-
# Generated by Django 1.9.7 on 2016-08-01 16:12
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone
import munch.apps.domains.fields
import munch.core.utils.models
import munch.core.utils.validators


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('users', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='SendingDomain',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=200, validators=[munch.core.utils.validators.DomainNameValidator()], verbose_name='name')),
                ('creation_date', models.DateTimeField(default=django.utils.timezone.now)),
                ('update_date', models.DateTimeField(blank=True)),
                ('dkim_status', munch.apps.domains.fields.DomainCheckField(choices=[('ok', 'Configured'), ('ko', 'Not configured'), ('bad', 'Badly configured'), ('pending', 'Checking'), ('unknown', 'Unknown')], default='unknown', max_length=10, verbose_name='DKIM status')),
                ('dkim_status_date', models.DateTimeField(blank=True, null=True, verbose_name='DKIM status last change')),
                ('app_domain', models.CharField(blank=True, default='', help_text='Domain to use for application links generation', max_length=200, validators=[munch.core.utils.validators.DomainNameValidator()], verbose_name='Custom application domain')),
                ('app_domain_status', munch.apps.domains.fields.DomainCheckField(choices=[('ok', 'Configured'), ('ko', 'Not configured'), ('bad', 'Badly configured'), ('pending', 'Checking'), ('unknown', 'Unknown')], default='unknown', max_length=10, verbose_name='App domain status')),
                ('app_domain_status_date', models.DateTimeField(blank=True, null=True, verbose_name='App domain status last change')),
                ('alt_organizations', models.ManyToManyField(blank=True, to='users.Organization')),
                ('organization', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='domains', to='users.Organization', verbose_name='organization')),
            ],
            options={
                'verbose_name': 'sending domain',
                'verbose_name_plural': 'sending domains',
                'default_permissions': ('view', 'add', 'change', 'delete', 'view_mine', 'change_mine', 'delete_mine', 'view_organizations', 'change_organizations', 'delete_organizations'),
                'abstract': False,
            },
            bases=(models.Model, munch.core.utils.models.OwnedModelMixin),
        ),
        migrations.AlterUniqueTogether(
            name='sendingdomain',
            unique_together=set([('name', 'organization')]),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.9.3 on 2016-04-04 12:14
from __future__ import unicode_literals

from django.conf import settings
import django.contrib.postgres.fields.hstore
import django.contrib.postgres.fields.jsonb
import django.core.validators
from django.db import migrations, models
import django.db.models.deletion
import munch.apps.contacts.validators
import munch.core.utils.models
import uuid


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='CollectedContact',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('properties', django.contrib.postgres.fields.hstore.HStoreField(default={}, verbose_name='propriétés')),
                ('address', models.EmailField(max_length=254, verbose_name='adresse e-mail')),
                ('creation_date', models.DateTimeField(auto_now_add=True, verbose_name='date de création')),
                ('update_date', models.DateTimeField(auto_now=True, verbose_name='date de modification')),
                ('uuid', models.UUIDField(default=uuid.uuid4, editable=False)),
                ('status', models.CharField(choices=[('pending', 'en attente'), ('bounced', 'échec bounce'), ('expired', 'expiré'), ('ok', 'ok'), ('consumed', 'consommé')], default='pending', max_length=50, verbose_name='statut')),
                ('subscription_ip', models.GenericIPAddressField(default='127.0.0.1', verbose_name='adresse IP d’envoi')),
            ],
            options={
                'verbose_name': 'adresse collectée',
                'verbose_name_plural': 'adresses collectées',
            },
            bases=(models.Model, munch.core.utils.models.OwnedModelMixin),
        ),
        migrations.CreateModel(
            name='Contact',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('properties', django.contrib.postgres.fields.hstore.HStoreField(default={}, verbose_name='propriétés')),
                ('address', models.EmailField(max_length=254, verbose_name='adresse e-mail')),
                ('creation_date', models.DateTimeField(auto_now_add=True, verbose_name='date de création')),
                ('update_date', models.DateTimeField(auto_now=True, verbose_name='date de modification')),
                ('uuid', models.UUIDField(default=uuid.uuid4, editable=False)),
                ('status', models.CharField(choices=[('pending', 'en attente'), ('bounced', 'échec bounce'), ('expired', 'expiré'), ('ok', 'ok'), ('consumed', 'consommé')], default='pending', max_length=50, verbose_name='statut')),
                ('subscription_ip', models.GenericIPAddressField(default='127.0.0.1', verbose_name='adresse IP d’envoi')),
            ],
            bases=(models.Model, munch.core.utils.models.OwnedModelMixin),
        ),
        migrations.CreateModel(
            name='ContactList',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('properties', django.contrib.postgres.fields.hstore.HStoreField(default={}, verbose_name='propriétés')),
                ('source_type', models.CharField(blank=True, max_length=100, validators=[django.core.validators.RegexValidator(message='source should be composed of lowercase letters and dashes', regex='^[a-z-]*$')])),
                ('source_ref', models.CharField(blank=True, max_length=200)),
                ('uuid', models.UUIDField(default=uuid.uuid4, editable=False)),
                ('name', models.CharField(max_length=100, verbose_name='nom')),
                ('contact_fields', django.contrib.postgres.fields.jsonb.JSONField(default=[{'name': 'Prénom', 'required': False, 'type': 'Char'}, {'name': 'Nom', 'required': False, 'type': 'Char'}], validators=[munch.apps.contacts.validators.properties_schema_validator], verbose_name='champs des contacts')),
                ('author', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL, verbose_name='auteur')),
            ],
            options={
                'verbose_name': 'liste de contacts',
                'verbose_name_plural': 'listes de contacts',
            },
            bases=(models.Model, munch.core.utils.models.OwnedModelMixin),
        ),
        migrations.CreateModel(
            name='ContactListPolicy',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=100, verbose_name='nom')),
                ('description', models.TextField(verbose_name='description')),
            ],
            options={
                'verbose_name': 'politique de file',
                'verbose_name_plural': 'politiques de files',
            },
        ),
        migrations.CreateModel(
            name='ContactListPolicyAttribution',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('contact_list', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='contacts.ContactList')),
                ('policy', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='contacts.ContactListPolicy')),
            ],
        ),
        migrations.CreateModel(
            name='ContactQueue',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('properties', django.contrib.postgres.fields.hstore.HStoreField(default={}, verbose_name='propriétés')),
                ('source_type', models.CharField(blank=True, max_length=100, validators=[django.core.validators.RegexValidator(message='source should be composed of lowercase letters and dashes', regex='^[a-z-]*$')])),
                ('source_ref', models.CharField(blank=True, max_length=200)),
                ('uuid', models.UUIDField(default=uuid.uuid4, editable=False)),
                ('contact_fields', django.contrib.postgres.fields.jsonb.JSONField(default=[], validators=[munch.apps.contacts.validators.properties_schema_validator], verbose_name='champs des contacts')),
                ('author', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL, verbose_name='auteur')),
            ],
            options={
                'abstract': False,
            },
            bases=(models.Model, munch.core.utils.models.OwnedModelMixin),
        ),
        migrations.CreateModel(
            name='ContactQueuePolicyAttribution',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('contact_queue', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='contacts.ContactQueue')),
                ('policy', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='contacts.ContactListPolicy')),
            ],
        ),
        migrations.AddField(
            model_name='contactqueue',
            name='policies',
            field=models.ManyToManyField(through='contacts.ContactQueuePolicyAttribution', to='contacts.ContactListPolicy', verbose_name='politiques de la file'),
        ),
        migrations.AddField(
            model_name='contactlist',
            name='policies',
            field=models.ManyToManyField(through='contacts.ContactListPolicyAttribution', to='contacts.ContactListPolicy', verbose_name='politiques de la file'),
        ),
        migrations.AddField(
            model_name='contact',
            name='contact_list',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='contacts', to='contacts.ContactList'),
        ),
        migrations.AddField(
            model_name='collectedcontact',
            name='contact_queue',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='collected_contacts', to='contacts.ContactQueue'),
        ),
        migrations.AlterUniqueTogether(
            name='contactqueuepolicyattribution',
            unique_together=set([('policy', 'contact_queue')]),
        ),
        migrations.AlterUniqueTogether(
            name='contactlistpolicyattribution',
            unique_together=set([('policy', 'contact_list')]),
        ),
        migrations.AlterUniqueTogether(
            name='contactlist',
            unique_together=set([('author', 'name')]),
        ),
        migrations.AlterUniqueTogether(
            name='contact',
            unique_together=set([('contact_list', 'address')]),
        ),
        migrations.AlterUniqueTogether(
            name='collectedcontact',
            unique_together=set([('contact_queue', 'address')]),
        ),
    ]



from django.db import models
from django.db.models import Count
from django.conf import settings
from django.utils import timezone
from django.utils.translation import ugettext_lazy as _
from django.template.loader import render_to_string

from munch.core.utils.models import AbstractOwnedModel
from munch.core.utils.managers import OwnedModelQuerySet
from munch.core.mail.utils.emails import NotificationMessage
from munch.apps.users.models import MunchUser


class OptOutManager(models.Manager):
    def create_or_update(
            self, identifier, address, origin,
            author=None, category=None, creation_date=None):
        """
        OptOuts from a Mail

        No matter if it already have a previous optout or not : create one or
        update the existing.
        """
        from .models import OptOut

        creation_date = creation_date or timezone.now()
        obj, created = OptOut.objects.get_or_create(
            identifier=identifier, address=address,
            defaults={
                'author': author, 'category': category,
                'origin': origin, 'creation_date': creation_date})
        if not created:
            obj.origin = OptOut.BY_WEB
            obj.creation_date = timezone.now()
            obj.save()


class OptOutQueryset(OwnedModelQuerySet):
    def count_by_origin(self, with_total=False):
        """
        Returns a dict with count for each optout origin

        It lists even optout origins with zero occurence for that queryset
        """
        from .models import OptOut

        d = {i: 0 for i, _ in OptOut._meta.get_field('origin').choices}
        d.update(self.values_list('origin').annotate(count=Count('origin')))
        if with_total:
            d['total'] = sum(d.values())
        return d

    def for_email(self, address):
        return self.filter(address=address).first()


class OptOut(AbstractOwnedModel):
    BY_MAIL = 'mail'
    BY_WEB = 'web'
    BY_FBL = 'feedback-loop'
    BY_BOUNCE = 'bounce'
    BY_ABUSE = 'abuse'
    BY_API = 'api'

    identifier = models.CharField(
        max_length=150, db_index=True,
        unique=True, verbose_name=_('identifier'))
    address = models.EmailField(verbose_name=_('mail'))
    creation_date = models.DateTimeField(default=timezone.now)
    origin = models.CharField(
        _('origine'), max_length=20, choices=(
            (BY_MAIL, _('Email')),
            (BY_WEB, _('Web link')),
            (BY_FBL, _('Detected as spam')),
            (BY_ABUSE, _("By abuse report")),
            (BY_BOUNCE, _('Too much delivering errors'))))
    author = models.ForeignKey(
        MunchUser, verbose_name=_('author'), null=True, blank=True)
    category = models.ForeignKey(
        'core.Category', verbose_name=_('category'), null=True, blank=True)

    objects = OptOutManager.from_queryset(OptOutQueryset)()

    class Meta(AbstractOwnedModel.Meta):
        verbose_name = _('optout')
        verbose_name_plural = _('optouts')

    owner_path = 'author__organization'
    author_path = 'author'

    def __str__(self):
        return self.identifier

    def get_organization(self):
        if self.author:
            return self.author.organization

    def notify_new_optout(self):
        from munch.core.utils import get_mail_by_identifier

        if not self.author:
            return

        organization = self.author.organization

        if organization.settings.notify_optouts and \
                organization.contact_email:
            mail = get_mail_by_identifier(self.identifier)
            message = mail.message
            origin = self.get_origin_display()

            message = NotificationMessage(
                subject=render_to_string(
                    'optouts/optout_notification_email_subject.txt', {
                        'product_name': settings.PRODUCT_NAME,
                        'address': self.address,
                        'origin': self.origin}).strip(),
                template='optouts/optout_notification_email.txt',
                render_context={
                    'to': self.address,
                    'origin': origin,
                    'message': message},
                to=organization.contact_email,
            )
            message.add_html_part_if_exists(
                'optouts/optout_notification_email.html')
            message.send()

    def save(self, *args, **kwargs):
        super().save(*args, **kwargs)
        self.notify_new_optout()



from django.shortcuts import render
from django.urls import reverse
from django.http import HttpResponseRedirect
from django.http import HttpResponseNotFound

from munch.core.utils import get_mail_by_identifier
from munch.apps.optouts.models import OptOut
from munch.apps.campaigns.models import PreviewMail

from .forms import AbuseNotificationForm


def abuse_report(request, identifier):
    is_test_mail = identifier.startswith('preview')
    if is_test_mail:
        mail = PreviewMail.objects.get(identifier=identifier)
    else:
        mail = get_mail_by_identifier(identifier=identifier, must_raise=False)
        if not mail:
            return HttpResponseNotFound()

    if request.method == 'POST':
        form = AbuseNotificationForm(request.POST)
        if is_test_mail:
            return HttpResponseRedirect(reverse(
                'abuse-report-thanks', kwargs={'identifier': identifier}))
        else:
            if form.is_valid():
                form.save()
                OptOut.objects.create_or_update(
                    author=mail.message.author,
                    category=mail.get_category(),
                    identifier=mail.identifier,
                    address=mail.recipient,
                    origin=OptOut.BY_ABUSE)
                return HttpResponseRedirect(reverse(
                    'abuse-report-thanks', kwargs={'identifier': identifier}))
    else:
        form = AbuseNotificationForm(initial={'mail': identifier})

    return render(
        request, 'abuse/report_abuse.html',
        {'form': form, 'mail': mail, 'is_test_mail': is_test_mail})


def abuse_report_thanks(request, identifier):
    is_test_mail = identifier.startswith('preview')
    if is_test_mail:
        mail = PreviewMail.objects.get(identifier=identifier)
    else:
        mail = get_mail_by_identifier(identifier=identifier, must_raise=False)
        if not mail:
            return HttpResponseNotFound()

    return render(
        request, 'abuse/report_thanks.html',
        {'mail': mail, 'is_test_mail': is_test_mail})



# -*- coding: utf-8 -*-
# Generated by Django 1.10.1 on 2016-09-23 13:19
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('abuse', '0003_set_default_for_comments'),
    ]

    operations = [
        migrations.AlterField(
            model_name='abusenotification',
            name='comments',
            field=models.TextField(verbose_name='comments'),
        ),
    ]



from rest_framework import permissions


class IsOwnerOrAdmin(permissions.BasePermission):
    def has_permission(self, request, view):
        if request.method in ['HEAD', 'OPTIONS']:
            return True
        return (
            request.user and request.user.is_authenticated() and (
                request.user.is_admin or request.user.organization))

    def has_object_permission(self, request, view, obj):
        return ((request.user.is_authenticated()) and
                ((request.user.is_admin) or (
                    obj.get_owner() == request.user.organization)))


class IsAuthenticatedOrPreflight(permissions.IsAuthenticated):
    def has_permission(self, request, view):
        if request.method == 'OPTIONS':
            return True
        return super().has_permission(request, view)


class MunchResourcePermission:
    app_name = 'campaigns'
    model_name = 'mail'

    @classmethod
    def mk_class(cls, model_class):
        class klass(cls):
            app_name = model_class._meta.app_label
            model_name = model_class._meta.model_name
        klass.__name__ = '{}Permission'.format(model_class.__name__)
        return klass

    # define a permission mapping HTTP methods
    verbs_permissions = {
        'GET': 'view',
        'HEAD': 'view',
        'POST': 'add',
        'PUT': 'change',
        'PATCH': 'change',
        'DELETE': 'delete'
    }

    def mk_perms(self, verb):
        """ Builds permissions names """
        return (
            self._mk_user_perm(verb),
            self._mk_organization_perm(verb),
            self._mk_main_perm(verb))

    def _mk_main_perm(self, verb):
        return '{}.{}_{}'.format(self.app_name, verb, self.model_name)

    def _mk_user_perm(self, verb):
        return '{}.{}_mine_{}'.format(self.app_name, verb, self.model_name)

    def _mk_organization_perm(self, verb):
        return '{}.{}_organizations_{}'.format(
            self.app_name, verb, self.model_name)

    def has_any_perm(self, user, perms):
        return any(user.has_perm(p) for p in perms)

    def has_permission(self, request, view):
        if request.method in ['HEAD', 'OPTIONS']:
            return True
        return self.has_any_perm(
            request.user,
            self.mk_perms(self.verbs_permissions[request.method]))

    def can_user_do(self, user, obj, verb):
        """ High-level object permissions checking

        :type user: MunchUser
        :type obj: the object we are interested in
        :param verb: 'add', 'change', 'delete' or 'view'

        :rtype: boolean
        """
        # if no author/owner, that's public resource
        # that can only be viewed
        if ((obj.defines_author() and obj.get_author() is None) or (
                obj.get_owner() is None)):
            return verb == 'view'
        elif user.has_perm(self._mk_main_perm(verb)):
            return True
        elif user.has_perm(self._mk_organization_perm(verb)):
            return user.organization == obj.get_owner()
        elif user.has_perm(self._mk_user_perm(verb)):
            return user == obj.get_author()
        return False

    def has_object_permission(self, request, view, obj):
        if not request.user.is_authenticated():
            return False
        if request.method == 'OPTIONS':
            return True
        else:
            perm_verb = self.verbs_permissions[request.method]
            return self.can_user_do(request.user, obj, perm_verb)

    def filter_queryset(self, qs, request):
        """ Filters queryset according to request and logged-in user perms
        """
        if request.user.is_authenticated():
            if request.user.has_perm(self._mk_main_perm('view')):
                return qs
            elif request.user.has_perm(self._mk_organization_perm('view')):
                return qs.from_owner(request.user.organization)
            elif request.user.has_perm(self._mk_user_perm('view')):
                return qs.from_author(request.user)

        return qs.none()



from django.db import transaction
from django.db.models import Prefetch
from django.db.utils import IntegrityError
from rest_framework import status
from rest_framework import viewsets
from rest_framework.response import Response
from rest_framework.settings import api_settings
from rest_framework.decorators import list_route
from rest_framework.decorators import detail_route
from rest_framework_bulk.mixins import BulkCreateModelMixin

from munch.core.utils.views import NestedView
from munch.core.utils.views import NestedViewMixin
from munch.core.utils.views import MunchModelViewSetMixin
from munch.core.utils.views import OrganizationOwnedViewSetMixin

from ...models import PROP_TYPES
from ...models import Contact
from ...models import ContactList
from ...models import ContactQueue
from ...models import ContactListPolicy
from .serializers import ContactSerializer
from .serializers import ContactListSerializer
from .serializers import NestedContactSerializer
from .serializers import ContactListListSerializer
from .permissions import ContactListMergePermission
from .serializers import QueuePolicySerializer
from .serializers import ContactQueueSerializer
from .serializers import ContactQueueDetailSerializer


class ContactListViewSet(
        OrganizationOwnedViewSetMixin,
        MunchModelViewSetMixin,
        viewsets.ModelViewSet):
    model = ContactList
    serializer_class = ContactListSerializer
    queryset = ContactList.objects.all()


class ContactViewSet(
        OrganizationOwnedViewSetMixin,
        MunchModelViewSetMixin,
        BulkCreateModelMixin,
        viewsets.ModelViewSet):
    model = Contact
    serializer_class = ContactSerializer

    def get_queryset(self):
        return Contact.objects.all().prefetch_related(Prefetch(
            'contact_list',
            queryset=ContactList.objects.all().only('contact_fields')))

    def create(self, request, *args, **kwargs):
        bulk = isinstance(request.data, list)

        if not bulk:
            return super(BulkCreateModelMixin, self).create(
                request, *args, **kwargs)

        else:

            serializer = self.get_serializer(data=request.data, many=True)
            has_errors = serializer.is_partly_invalid()

            if not has_errors:
                self.perform_bulk_create(serializer)
                return Response(status=status.HTTP_201_CREATED)
            else:
                return Response(
                    data=serializer._partial_errors,
                    status=status.HTTP_400_BAD_REQUEST)


class ContactListContacts(ContactViewSet, NestedViewMixin):
    """Access and import *contacts* to a given *list*.

    # CSV import

    - set correct content-type : `Content-type: text/csv; charset=utf-8`
    - your CSV first row should be header (field names)


    `address` field will go to `address` field on contact resource, all the
       other fields will go in the *properties* dict.

    If your CSV file contains more fields than you need, you can include
    only some with `fields` url querystring, (i.e: `?fields=[foo,bar]`
    to import only `foo` and `bar` fields), others will be discarded.
    """
    serializer_class = NestedContactSerializer
    parent_model = ContactList

    def get_parent_object(self, pk=None):
        if pk is None:
            pk = self.kwargs['contact_list_pk']
        return super().get_parent_object(pk)

    def get_queryset(self):
        contact_list = self.get_parent_object()
        return contact_list.contacts.filter(contact_list=contact_list)

    def get_serializer_context(self):
        context = super().get_serializer_context()
        context['kwargs'] = self.kwargs
        return context


class ContactListMergeView(NestedView):
    """ Merge several contact lists into one

    The *master list* that will receive the others contents is the
    one pointed by URL.

    Request payload is an array of *lists* URL to be merged in
    the *master list*.

    *e.g:*

        POST /v1/contact/lists/1/merge/
        ['/v1/contacts/lists/2/', '/v1/contacts/lists/3/']

    The contact lists in arguments will then be deleted.

    """
    parent_model = ContactList
    permission_classes = [ContactListMergePermission]

    def post(self, request, contact_list_pk):
        master_list = self.get_parent_object(contact_list_pk)

        serializer = ContactListListSerializer(
            data={'contact_lists': request.data},
            master_list=master_list,
            context={'request': request},)

        if serializer.is_valid():
            included_lists = serializer.validated_data['contact_lists']
            # Optimistic merge: we try first and de-duplicate if the re is any
            # error on update
            try:
                with transaction.atomic():
                    Contact.objects.filter(contact_list__in=included_lists)\
                                   .update(contact_list=master_list)
            except IntegrityError:
                # get the doubles
                double_addrs = (
                    set(Contact.objects.filter(
                        contact_list__in=included_lists).values_list(
                            'address', flat=True)) &
                    set(Contact.objects.filter(contact_list=master_list)
                        .values_list('address', flat=True))
                )
                Contact.objects.filter(contact_list__in=included_lists)\
                               .exclude(address__in=double_addrs)\
                               .update(contact_list=master_list)

            for i in included_lists:
                i.delete()
            master_list_serializer = ContactListSerializer(
                master_list, context={'request': request})

            return Response(
                data=master_list_serializer.data, status=status.HTTP_200_OK)

        else:
            errors = {api_settings.NON_FIELD_ERRORS_KEY: []}

            # put every error in non_field_errors
            if api_settings.NON_FIELD_ERRORS_KEY not in errors:
                errors[api_settings.NON_FIELD_ERRORS_KEY] = []
            for k, v in serializer.errors.items():
                errors[api_settings.NON_FIELD_ERRORS_KEY] += v

            return Response(data=errors, status=status.HTTP_400_BAD_REQUEST)


class ContactQueueViewSet(
        OrganizationOwnedViewSetMixin,
        MunchModelViewSetMixin,
        viewsets.ModelViewSet):
    model = ContactQueue
    serializer_class = ContactQueueSerializer
    queryset = ContactQueue.objects.all()

    @detail_route(methods=['post', 'get'])
    def consume(self, request, pk):
        queue = self.get_object()
        if 'status' in request.GET:
            queue.contacts = queue.consume(status=request.GET['status'])
        else:
            queue.contacts = queue.consume()

        serializer = ContactQueueDetailSerializer(
            queue, context={'request': self.request})
        return Response(serializer.data)

    @list_route(methods=['get'])
    def fields(self, request):
        return Response(PROP_TYPES)


class ContactListPolicyViewSet(MunchModelViewSetMixin, viewsets.ModelViewSet):
    model = ContactListPolicy
    serializer_class = QueuePolicySerializer
    queryset = ContactListPolicy.objects.all()



from django.test import TestCase

from munch.core.tests.factories import CategoryFactory
from munch.apps.users.tests.factories import UserFactory
from munch.apps.optouts.tests.factories import OptOutFactory

from ..models import OptOut
from .factories import MailFactory
from .factories import MessageFactory


class OptOuts(TestCase):
    def setUp(self):
        self.user = UserFactory()

    def test_optout_no_send(self):
        """
        Check that previous optouts from same
        category are honored in current category.
        """
        # A prev_message, in same category, with an optout
        category = CategoryFactory(author=self.user)
        old_message = MessageFactory(author=self.user, category=category)
        dark_mail = MailFactory(message=old_message)
        OptOutFactory(
            author=dark_mail.message.author,
            category=dark_mail.message.category,
            identifier=dark_mail.identifier,
            address=dark_mail.recipient,
            origin=OptOut.BY_WEB)

        # A message, in same category
        new_message = MessageFactory(author=self.user, category=category)
        MailFactory(message=new_message, recipient=dark_mail.recipient)
        good_mail = MailFactory(message=new_message)

        self.assertEqual(
            list(new_message.willsend_addresses()), [good_mail.recipient])
        self.assertEqual(
            list(new_message.willnotsend_addresses()),
            [dark_mail.recipient])

    def test_optout_send(self):
        """
        Check that previous optouts from different
        category are ignored in current category.
        """
        # A prev_message, in different category, with an optout
        old_category = CategoryFactory(author=self.user)
        old_message = MessageFactory(author=self.user, category=old_category)
        dark_mail = MailFactory(message=old_message)
        OptOutFactory(
            author=dark_mail.message.author,
            category=dark_mail.message.category,
            identifier=dark_mail.identifier,
            address=dark_mail.recipient,
            origin=OptOut.BY_WEB)

        # A message, in same category
        new_category = CategoryFactory(author=self.user)
        new_message = MessageFactory(author=self.user, category=new_category)
        MailFactory(message=new_message, recipient=dark_mail.recipient)
        good_mail = MailFactory(message=new_message)

        self.assertEqual(
            set(list(new_message.willsend_addresses())),
            set([dark_mail.recipient, good_mail.recipient]))
        self.assertEqual(list(new_message.willnotsend_addresses()), [])

    def test_optout_count(self):
        category = CategoryFactory(author=self.user)
        message = MessageFactory(author=self.user, category=category)

        cases = [
            ('1@example.com', OptOut.BY_WEB),
            ('2@example.com', OptOut.BY_FBL),
            ('4@example.com', OptOut.BY_BOUNCE),
            ('3@example.com', OptOut.BY_ABUSE),
            ('5@example.com', OptOut.BY_ABUSE)]
        for addr, origin in cases:
            mail = MailFactory(message=message, recipient=addr)
            OptOutFactory(
                identifier=mail.identifier,
                address=mail.recipient,
                origin=origin)

        res = OptOut.objects.count_by_origin()
        self.assertEqual(
            res,
            {'mail': 0, 'abuse': 2, 'feedback-loop': 1, 'bounce': 1, 'web': 1})



from django.conf import settings


def set_unsubscribe_url(
        text,
        unsubscribe_url=None,
        no_unsubscribe_placehoder_must_raise=False, **kwargs):
    """
    :param text_input: can be any text, including html
    """
    placeholder = settings.OPTOUTS['UNSUBSCRIBE_PLACEHOLDER']
    if unsubscribe_url:
        if placeholder in text:
            return text.replace(placeholder, unsubscribe_url)
        else:
            if no_unsubscribe_placehoder_must_raise:
                raise Exception('Cannot find unsubscribe placeholder')

    return text



from django.conf import settings
from django.test import TestCase
from django.utils import timezone

from slimta.envelope import Envelope
from slimta.relay import PermanentRelayError
from slimta.queue import QueueError
from faker import Factory as FakerFactory

from munch.core.models import Category
# from munch.core.utils.tests import temporary_settings
# from munch.core.tests.factories import CategoryFactory
from munch.apps.users.tests.factories import UserFactory
from munch.apps.users.tests.factories import SmtpApplicationFactory
from munch.apps.domains.tests.factories import SendingDomainFactory
from munch.apps.optouts.models import OptOut
from munch.apps.optouts.tests.factories import OptOutFactory

from ..models import Mail
from ..models import MailBatch
from ..policies.relay import headers as headers_policy
from ..policies.queue import bounces
from ..policies.queue import identifier
from ..policies.queue import store_mail
from ..policies.queue import sending_domain

faker = FakerFactory.create()


class PolicyCase(TestCase):
    def mk_envelope(self, data, sender=None, recipients=None):
        env = Envelope(sender=sender)
        env.parse(data)
        if not recipients:
            recipients = ['root@localhost']
        env.recipients = recipients
        return env


class TestStoreHeaders(PolicyCase):
    def setUp(self):
        self.smtp_application = SmtpApplicationFactory()
        self.domain = SendingDomainFactory(
            name='example.com',
            organization=self.smtp_application.author.organization)
        self.user = UserFactory(
            identifier='foo@example.com', last_login=timezone.now())

    def test_store_mail_empty(self):
        env = self.mk_envelope(b'')
        env.client = {'auth': (self.smtp_application.username, None)}
        env.user = self.user

        with self.assertRaises(PermanentRelayError):
            store_mail.Store().apply(env)

    def test_store_mail_minimal(self):
        env = self.mk_envelope(
            b"From: test-from@example.com\nSubject: foo\nTo: foo@bar",
            sender="test@example.com", recipients=['test-to@example.com'])
        env.client = {'auth': (self.smtp_application.username, None)}
        env.user = self.user

        identifier.Add().apply(env)
        bounces.Check().apply(env)
        sending_domain.Check().apply(env)
        store_mail.Store().apply(env)
        mail = Mail.objects.get(
            identifier=env.headers.get(settings.TRANSACTIONAL[
                'X_MESSAGE_ID_HEADER']))
        self.assertEqual(
            mail.headers, {
                'Subject': 'foo',
                'To': 'foo@bar',
                'From': "test-from@example.com",
                settings.TRANSACTIONAL[
                    'X_MESSAGE_ID_HEADER']: env.headers.get(
                        settings.TRANSACTIONAL['X_MESSAGE_ID_HEADER']),
                settings.TRANSACTIONAL[
                    'X_USER_ID_HEADER']: str(self.user.pk)})
        self.assertEqual(mail.sender, 'test@example.com')
        self.assertEqual(MailBatch.objects.count(), 0)
        self.assertEqual(Category.objects.count(), 0)

    def test_store_mail_with_batch(self):
        headers = (
            "From: test-from@example.com\n"
            "Subject: foo\nTo: foo@bar\n{}: foo").format(
                settings.TRANSACTIONAL['X_MAIL_BATCH_HEADER'])
        env = self.mk_envelope(
            headers.encode('utf-8'),
            sender="test@example.com", recipients=['test-to@example.com'])
        env.client = {'auth': (self.smtp_application.username, None)}
        env.user = self.user

        identifier.Add().apply(env)
        bounces.Check().apply(env)
        sending_domain.Check().apply(env)
        store_mail.Store().apply(env)
        mail = Mail.objects.get(
            identifier=env.headers.get(settings.TRANSACTIONAL[
                'X_MESSAGE_ID_HEADER']))

        self.assertEqual(mail.batch.name, 'foo')

        headers = (
            "From: test-from@example.com\n"
            "Subject: foo\nTo: foo@bar\n{}: foo").format(
                settings.TRANSACTIONAL['X_MAIL_BATCH_HEADER'])
        env = self.mk_envelope(
            headers.encode('utf-8'),
            sender="test@example.com", recipients=['test-to@example.com'])
        env.client = {'auth': (self.smtp_application.username, None)}
        env.user = self.user

        identifier.Add().apply(env)
        bounces.Check().apply(env)
        sending_domain.Check().apply(env)
        store_mail.Store().apply(env)

        self.assertEqual(MailBatch.objects.count(), 1)
        self.assertEqual(Category.objects.count(), 0)

    def test_store_mail_another_with_category(self):
        headers = (
            "From: test-from@example.com\n"
            "Subject: foo\nTo: foo@bar\n{}: foo").format(
                settings.TRANSACTIONAL['X_MAIL_BATCH_HEADER'])
        env = self.mk_envelope(
            headers.encode('utf-8'),
            sender="test@example.com", recipients=['test-to@example.com'])
        env.client = {'auth': (self.smtp_application.username, None)}
        env.user = self.user

        identifier.Add().apply(env)
        bounces.Check().apply(env)
        sending_domain.Check().apply(env)
        store_mail.Store().apply(env)
        mail = Mail.objects.get(
            identifier=env.headers.get(settings.TRANSACTIONAL[
                'X_MESSAGE_ID_HEADER']))

        self.assertEqual(mail.batch.name, 'foo')

        headers = (
            "From: test-from@example.com\n"
            "Subject: foo\nTo: foo@bar\n{}: foo\n{}: foo-cat").format(
                settings.TRANSACTIONAL['X_MAIL_BATCH_HEADER'],
                settings.TRANSACTIONAL['X_MAIL_BATCH_CATEGORY_HEADER'])
        env = self.mk_envelope(
            headers.encode('utf-8'),
            sender="test@example.com", recipients=['test-to@example.com'])
        env.client = {'auth': (self.smtp_application.username, None)}
        env.user = self.user

        identifier.Add().apply(env)
        bounces.Check().apply(env)
        sending_domain.Check().apply(env)
        store_mail.Store().apply(env)

        self.assertEqual(MailBatch.objects.count(), 1)
        self.assertEqual(Category.objects.count(), 1)

    def test_store_mail_category_scope(self):
        # Create first mail, batch and category
        headers = (
            "From: test-from@example.com\n"
            "Subject: foo\nTo: foo@bar\n{}: foo\n{}: foo-cat").format(
                settings.TRANSACTIONAL['X_MAIL_BATCH_HEADER'],
                settings.TRANSACTIONAL['X_MAIL_BATCH_CATEGORY_HEADER'])
        env = self.mk_envelope(
            headers.encode('utf-8'),
            sender="test@example.com", recipients=['test-to@example.com'])
        env.client = {'auth': (self.smtp_application.username, None)}
        env.user = self.user

        identifier.Add().apply(env)
        bounces.Check().apply(env)
        sending_domain.Check().apply(env)
        store_mail.Store().apply(env)
        # mail = Mail.objects.get(
        #     identifier=env.headers.get(settings.TRANSACTIONAL[
        #         'X_MESSAGE_ID_HEADER']))

        # Same batch name and category with another user
        another_smtp_application = SmtpApplicationFactory()
        SendingDomainFactory(
            name='example.com',
            organization=another_smtp_application.author.organization)

        headers = (
            "From: test-from@example.com\n"
            "Subject: foo\nTo: foo@bar\n{}: foo\n{}: foo-cat").format(
                settings.TRANSACTIONAL['X_MAIL_BATCH_HEADER'],
                settings.TRANSACTIONAL['X_MAIL_BATCH_CATEGORY_HEADER'])
        env = self.mk_envelope(
            headers.encode('utf-8'),
            sender="test@example.com", recipients=['test-to@example.com'])
        env.client = {'auth': (another_smtp_application.username, None)}
        env.user = self.user

        identifier.Add().apply(env)
        bounces.Check().apply(env)
        sending_domain.Check().apply(env)
        store_mail.Store().apply(env)
        # Mail.objects.get(
        #     identifier=env.headers.get(settings.TRANSACTIONAL[
        #         'X_MESSAGE_ID_HEADER']))

        self.assertEqual(Mail.objects.filter(
            author=self.smtp_application.author).count(), 1)
        self.assertEqual(Mail.objects.filter(
            author=another_smtp_application.author).count(), 1)
        self.assertEqual(Mail.objects.count(), 2)

        self.assertEqual(MailBatch.objects.filter(
            author=self.smtp_application.author).count(), 1)
        self.assertEqual(MailBatch.objects.filter(
            author=another_smtp_application.author).count(), 1)
        self.assertEqual(MailBatch.objects.count(), 2)

        self.assertEqual(Category.objects.filter(
            author=self.smtp_application.author).count(), 1)
        self.assertEqual(Category.objects.filter(
            author=another_smtp_application.author).count(), 1)
        self.assertEqual(Category.objects.count(), 2)

    def test_track_open_no_html(self):
        headers = (
            "From: test-from@example.com\n"
            "Subject: foo\nTo: foo@bar\n{}: true").format(
                settings.TRANSACTIONAL['X_MAIL_TRACK_OPEN_HEADER'])
        env = self.mk_envelope(
            headers.encode('utf-8') + '\n\nMy Message'.encode('utf-8'),
            sender="test@example.com", recipients=['test-to@example.com'])
        env.client = {'auth': (self.smtp_application.username, None)}
        env.user = self.user

        identifier.Add().apply(env)
        bounces.Check().apply(env)
        sending_domain.Check().apply(env)
        store_mail.Store().apply(env)
        headers, body = env.flatten()
        self.assertNotIn(
            'alt="" height="1" width="1" border="0" />', body.decode('utf-8'))
        self.assertNotIn(
            '/t/open/{}'.format(env.headers.get(settings.TRANSACTIONAL[
                'X_MESSAGE_ID_HEADER'])), body.decode('utf-8'))

    def test_track_open(self):
        headers = (
            'From: test-from@example.com\n'
            'Content-Type: multipart/alternative;\n'
            ' boundary="===============0445577956452755870=="\n'
            'Subject: foo\n'
            'To: foo@bar\n'
            '{}: true').format(
                settings.TRANSACTIONAL['X_MAIL_TRACK_OPEN_HEADER'])
        body = (
            '--===============0445577956452755870==\n'
            'Content-Type: text/plain; charset="us-ascii"\n'
            'MIME-Version: 1.0\n'
            'Content-Transfer-Encoding: 7bit\n'
            '\n'
            'My template in text to [1]\n'
            '\n'
            '[1]: http://google.it\n'
            '--===============0445577956452755870==\n'
            'Content-Type: text/html; charset="us-ascii"\n'
            'MIME-Version: 1.0\n'
            'Content-Transfer-Encoding: 7bit\n'
            '\n'
            '<a href="http://google.fr">Google!</a>\n'
            '<a href="http://google.com">Google COM</a>\n'
            '<strong>Strong jambon is strong</strong>\n'
            '--===============0445577956452755870==--\n')
        env = self.mk_envelope(
            headers.encode('utf-8') + body.encode('utf-8'),
            sender="test@example.com", recipients=['test-to@example.com'])
        env.client = {'auth': (self.smtp_application.username, None)}
        env.user = self.user

        identifier.Add().apply(env)
        bounces.Check().apply(env)
        sending_domain.Check().apply(env)
        store_mail.Store().apply(env)
        headers, body = env.flatten()
        self.assertIn(
            'alt="" height="1" width="1" border="0"', body.decode('utf-8'))
        self.assertIn(
            '/t/open/{}'.format(env.headers.get(settings.TRANSACTIONAL[
                'X_MESSAGE_ID_HEADER'])), body.decode('utf-8'))

    def test_track_clicks(self):
        headers = (
            'From: test-from@example.com\n'
            'Content-Type: multipart/alternative;\n'
            ' boundary="===============0445577956452755870=="\n'
            'Subject: foo\n'
            'To: foo@bar\n'
            '{}: true').format(
                settings.TRANSACTIONAL['X_MAIL_TRACK_CLICKS_HEADER'])
        body = (
            '--===============0445577956452755870==\n'
            'Content-Type: text/plain; charset="us-ascii"\n'
            'MIME-Version: 1.0\n'
            'Content-Transfer-Encoding: 7bit\n'
            '\n'
            'My template in text to [1]\n'
            '\n'
            '[1]: http://google.it\n'
            '--===============0445577956452755870==\n'
            'Content-Type: text/html; charset="us-ascii"\n'
            'MIME-Version: 1.0\n'
            'Content-Transfer-Encoding: 7bit\n'
            '\n'
            '<a href="http://google.fr">Google!</a>\n'
            '<a href="http://google.com">Google COM</a>\n'
            '<strong>Strong jambon is strong</strong>\n'
            '--===============0445577956452755870==--\n')
        env = self.mk_envelope(
            headers.encode('utf-8') + body.encode('utf-8'),
            sender="test@example.com", recipients=['test-to@example.com'])
        env.client = {'auth': (self.smtp_application.username, None)}
        env.user = self.user

        identifier.Add().apply(env)
        bounces.Check().apply(env)
        sending_domain.Check().apply(env)
        store_mail.Store().apply(env)
        headers, body = env.flatten()
        self.assertIn('/t/clicks/m/', body.decode('utf-8'))
        self.assertIn('http://google.it', body.decode('utf-8'))
        self.assertNotIn('http://google.fr', body.decode('utf-8'))
        self.assertNotIn('http://google.com', body.decode('utf-8'))

    def test_unsubscribe_link_no_placeholder(self):
        headers = (
            'From: test-from@example.com\n'
            'Content-Type: multipart/alternative;\n'
            ' boundary="===============0445577956452755870=="\n'
            'Subject: foo\n'
            'To: foo@bar\n'
            '{}: true').format(
                settings.TRANSACTIONAL['X_MAIL_UNSUBSCRIBE_HEADER'])
        body = (
            '--===============0445577956452755870==\n'
            'Content-Type: text/plain; charset="us-ascii"\n'
            'MIME-Version: 1.0\n'
            'Content-Transfer-Encoding: 7bit\n'
            '\n'
            'My template in text to [1]\n'
            '\n'
            '[1]: http://google.it\n'
            '--===============0445577956452755870==\n'
            'Content-Type: text/html; charset="us-ascii"\n'
            'MIME-Version: 1.0\n'
            'Content-Transfer-Encoding: 7bit\n'
            '\n'
            '<a href="http://google.fr">Google!</a>\n'
            '<a href="http://google.com">Google COM</a>\n'
            '<strong>Strong jambon is strong</strong>\n'
            '--===============0445577956452755870==--\n')
        env = self.mk_envelope(
            headers.encode('utf-8') + body.encode('utf-8'),
            sender="test@example.com", recipients=['test-to@example.com'])
        env.client = {'auth': (self.smtp_application.username, None)}
        env.user = self.user

        identifier.Add().apply(env)
        bounces.Check().apply(env)
        sending_domain.Check().apply(env)
        store_mail.Store().apply(env)
        headers, body = env.flatten()
        self.assertNotIn('/h/subscriptions/', body.decode('utf-8'))

    def test_unsubscribe_link(self):
        headers = (
            'From: test-from@example.com\n'
            'Content-Type: multipart/alternative;\n'
            ' boundary="===============0445577956452755870=="\n'
            'Subject: foo\n'
            'To: foo@bar\n'
            '{}: true').format(
                settings.TRANSACTIONAL['X_MAIL_UNSUBSCRIBE_HEADER'])
        body = (
            '--===============0445577956452755870==\n'
            'Content-Type: text/plain; charset="us-ascii"\n'
            'MIME-Version: 1.0\n'
            'Content-Transfer-Encoding: 7bit\n'
            '\n'
            'My template in text to [1]\n'
            'Unsub here: {}'
            '\n'
            '[1]: http://google.it\n'
            '--===============0445577956452755870==\n'
            'Content-Type: text/html; charset="us-ascii"\n'
            'MIME-Version: 1.0\n'
            'Content-Transfer-Encoding: 7bit\n'
            '\n'
            '<a href="http://google.fr">Google!</a>\n'
            '<a href="http://google.com">Google COM</a>\n'
            '<strong>Strong jambon is strong</strong>\n'
            '<a href="{}">Unsubscribe here</a>\n'
            '--===============0445577956452755870==--\n').format(
                settings.OPTOUTS['UNSUBSCRIBE_PLACEHOLDER'],
                settings.OPTOUTS['UNSUBSCRIBE_PLACEHOLDER'])
        env = self.mk_envelope(
            headers.encode('utf-8') + body.encode('utf-8'),
            sender="test@example.com", recipients=['test-to@example.com'])
        env.client = {'auth': (self.smtp_application.username, None)}
        env.user = self.user

        identifier.Add().apply(env)
        bounces.Check().apply(env)
        sending_domain.Check().apply(env)
        store_mail.Store().apply(env)
        headers, body = env.flatten()
        self.assertIn('/h/subscriptions/', body.decode('utf-8'))


class TestBounces(PolicyCase):
    def setUp(self):
        self.smtp_application = SmtpApplicationFactory()
        self.domain = SendingDomainFactory(
            name='example.com',
            organization=self.smtp_application.author.organization)
        self.user = UserFactory(
            identifier='foo@example.com', last_login=timezone.now())

    def test_clean_address(self):
        optout = OptOutFactory(origin=OptOut.BY_WEB)
        envelope = self.mk_envelope(b'', recipients=[optout.address])
        envelope.client = {'auth': (self.smtp_application.username, None)}
        envelope.user = self.user
        identifier.Add().apply(envelope)
        bounces.Check().apply(envelope)

    def test_bounce(self):
        optout = OptOutFactory(origin=OptOut.BY_BOUNCE)
        envelope = self.mk_envelope(b'', recipients=[optout.address])
        envelope.client = {'auth': (self.smtp_application.username, None)}
        envelope.user = self.user
        identifier.Add().apply(envelope)
        with self.assertRaises(QueueError):
            bounces.Check().apply(envelope)

    def test_optout_batch_no_category(self):
        headers = (
            'From: test-from@example.com\n'
            'Subject: foo\n'
            'To: foo@bar\n'
            '{}: test').format(
                settings.TRANSACTIONAL['X_MAIL_BATCH_HEADER'])
        recipient = faker.email()
        envelope = self.mk_envelope(
            headers.encode('utf-8'),
            sender=faker.email(), recipients=[recipient])
        envelope.client = {'auth': (self.smtp_application.username, None)}
        envelope.user = self.user
        identifier.Add().apply(envelope)
        bounces.Check().apply(envelope)
        sending_domain.Check().apply(envelope)
        store_mail.Store().apply(envelope)

        self.client.post(
            '/h/subscriptions/{}/optout/'.format(
                envelope.headers.get(
                    settings.TRANSACTIONAL['X_MESSAGE_ID_HEADER'])))

        with self.assertRaises(QueueError):
            identifier.Add().apply(envelope)
            bounces.Check().apply(envelope)

    def test_optout_batch_no_category_another_user(self):
        headers = (
            'From: test-from@example.com\n'
            'Subject: foo\n'
            'To: foo@bar\n'
            '{}: test').format(
                settings.TRANSACTIONAL['X_MAIL_BATCH_HEADER'])
        recipient = faker.email()
        envelope = self.mk_envelope(
            headers.encode('utf-8'),
            sender=faker.email(), recipients=[recipient])
        envelope.client = {'auth': (self.smtp_application.username, None)}
        envelope.user = self.user
        identifier.Add().apply(envelope)
        bounces.Check().apply(envelope)
        sending_domain.Check().apply(envelope)
        store_mail.Store().apply(envelope)

        self.client.post(
            '/h/subscriptions/{}/optout/'.format(
                envelope.headers.get(
                    settings.TRANSACTIONAL['X_MESSAGE_ID_HEADER'])))

        another_smtp_application = SmtpApplicationFactory()
        SendingDomainFactory(
            name='example.com',
            organization=another_smtp_application.author.organization)
        headers = (
            'From: test-from@example.com\n'
            'Subject: foo\n'
            'To: foo@bar\n'
            '{}: test').format(
                settings.TRANSACTIONAL['X_MAIL_BATCH_HEADER'])
        recipient = faker.email()
        envelope = self.mk_envelope(
            headers.encode('utf-8'),
            sender=faker.email(), recipients=[recipient])
        envelope.client = {'auth': (another_smtp_application.username, None)}
        envelope.user = self.user
        identifier.Add().apply(envelope)
        bounces.Check().apply(envelope)
        sending_domain.Check().apply(envelope)
        store_mail.Store().apply(envelope)

    def test_optout_batch_with_category(self):
        # First Optout on mail without category
        headers = (
            'From: test-from@example.com\n'
            'Subject: foo\n'
            'To: foo@bar\n'
            '{}: test').format(
                settings.TRANSACTIONAL['X_MAIL_BATCH_HEADER'])
        recipient = faker.email()
        envelope = self.mk_envelope(
            headers.encode('utf-8'),
            sender=faker.email(), recipients=[recipient])
        envelope.client = {'auth': (self.smtp_application.username, None)}
        envelope.user = self.user
        identifier.Add().apply(envelope)
        bounces.Check().apply(envelope)
        sending_domain.Check().apply(envelope)
        store_mail.Store().apply(envelope)

        self.client.post(
            '/h/subscriptions/{}/optout/'.format(
                envelope.headers.get(
                    settings.TRANSACTIONAL['X_MESSAGE_ID_HEADER'])))

        # Second one with same recipient but with a category
        headers = (
            'From: test-from@example.com\n'
            'Subject: foo\n'
            'To: foo@bar\n'
            '{}: test-category\n'
            '{}: test').format(
                settings.TRANSACTIONAL['X_MAIL_BATCH_CATEGORY_HEADER'],
                settings.TRANSACTIONAL['X_MAIL_BATCH_HEADER'])
        envelope = self.mk_envelope(
            headers.encode('utf-8'),
            sender=faker.email(), recipients=[recipient])
        envelope.client = {'auth': (self.smtp_application.username, None)}
        envelope.user = self.user
        identifier.Add().apply(envelope)
        bounces.Check().apply(envelope)
        sending_domain.Check().apply(envelope)
        store_mail.Store().apply(envelope)

        identifier.Add().apply(envelope)
        bounces.Check().apply(envelope)


class TestReturnPath(PolicyCase):
    def test_no_prev_returnpath(self):
        """
        X-Munch-HTTP-Return-Path: yes
        Return-Path: no

        Return-Path-Rewrite: yes
        """
        env = self.mk_envelope(b'X-Munch-HTTP-Return-Path: http://unittest')
        headers_policy.RewriteReturnPath().apply(env)
        headers, _ = env.flatten()
        self.assertIn('return-', env.sender)

    def test_no_prev_returnpath_without_http_returnpath(self):
        """
        X-Munch-HTTP-Return-Path: no
        Return-Path: no

        Return-Path: envelope.sender
        """
        env = self.mk_envelope(b'', sender="foo@bar.com")
        headers_policy.RewriteReturnPath().apply(env)
        headers, _ = env.flatten()
        self.assertIn('foo@bar.com', env.sender)

    def test_replace_returnpath(self):
        """
        X-Munch-HTTP-Return-Path: yes
        Return-Path: yes

        Return-Path-Rewrite: yes
        """
        env = self.mk_envelope(
            b"X-Munch-HTTP-Return-Path: http://unittest.example.com/ping",
            sender='rp-124@example.com')
        headers_policy.RewriteReturnPath().apply(env)
        headers, _ = env.flatten()

        self.assertRegex(env.sender, 'return-.*test.munch.example.com.*')

    def test_replace_returnpath_without_http_returnpath(self):
        """
        X-Munch-HTTP-Return-Path: no
        Return-Path: yes

        Return-Path-Rewrite: no
        """
        env = self.mk_envelope(b"Return-Path: rp-1234@example.com")
        headers_policy.RewriteReturnPath().apply(env)
        headers, _ = env.flatten()

        self.assertIn(
            'Return-Path: rp-1234@example.com', headers.decode('utf-8'))



from django.apps import AppConfig
from django.utils.translation import ugettext_lazy as _


class CoreApp(AppConfig):
    name = 'munch.core'
    verbose_name = _("Core")

    def ready(self):
        import munch.core.signals  # noqa
        import munch.core.api.v1.urls  # noqa



import random
from unittest.mock import patch

import PIL.Image
from django.test import TestCase
from django.core.files.temp import NamedTemporaryFile
from rest_framework.test import APIClient

from munch.apps.users.tests.factories import UserFactory
from munch.core.utils.tests import HttpBasicTestCaseMixin

from ...models import Image
from ...tests import ImageTestCaseMixin
from ...tests import mock_get_app_url


class ImageAPITestCase(TestCase, ImageTestCaseMixin):
    def setUp(self):
        self.api_prefix = 'v1'
        self.admin = UserFactory(
            is_admin=True, is_active=True, is_superuser=True)
        self.admin.organization.can_attach_files = False
        self.admin.organization.can_external_optout = False
        self.admin.organization.save()

        self.client = APIClient()
        self.client.login(
            identifier=self.admin.identifier, password='password')
        self.organization = self.admin.organization

    def test_api_listing(self):
        response = self.client.get('/{}/upload-store/images/'.format(
            self.api_prefix))
        self.assertEqual(405, response.status_code)

    def test_api_creation(self):
        file = self.create_random_image()

        response = self.client.post(
            '/{}/upload-store/images/'.format(self.api_prefix),
            {'file': file}, format='multipart')
        self.assertEqual(201, response.status_code)
        obj = Image.objects.last()
        self.assertRegexpMatches(
            response.data['file'], '{}\.png$'.format(obj.hash))

    def test_api_dup_creation(self):
        file = self.create_random_image()

        response = self.client.post(
            '/{}/upload-store/images/'.format(self.api_prefix),
            {'file': file}, format='multipart')
        self.assertEqual(201, response.status_code)

        file.file.seek(0)
        response = self.client.post(
            '/{}/upload-store/images/'.format(self.api_prefix),
            {'file': file}, format='multipart')
        self.assertEqual(201, response.status_code)

        file.file.seek(0)
        response = self.client.post(
            '/{}/upload-store/images/'.format(self.api_prefix),
            {'file': file, 'expiration': '10:00'}, format='multipart')
        self.assertEqual(201, response.status_code)

    def test_api_creation_with_width(self):
        file = self.create_random_image()

        response = self.client.post(
            '/{}/upload-store/images/'.format(self.api_prefix),
            {'file': file, 'width': 100, 'expiration': '10:00'},
            format='multipart')
        self.assertEqual(201, response.status_code)
        self.assertEqual(100, response.data['width'])
        self.assertEqual('00:10:00', response.data['expiration'])

        obj = Image.objects.last()
        image_file = PIL.Image.open(obj.file.file)
        self.assertEqual((100, 75), image_file.size)

    def test_api_creation_with_too_large_width(self):
        actual_file_width = 2
        asked_file_width = 10
        file = self.create_random_image(width=actual_file_width, height=2)

        response = self.client.post(
            '/{}/upload-store/images/'.format(self.api_prefix),
            {'file': file, 'width': asked_file_width, 'expiration': '10:00'},
            format='multipart')
        self.assertEqual(201, response.status_code)
        self.assertEqual(actual_file_width, response.data['width'])

    def test_api_retrieve(self):
        file = self.create_random_image()

        response = self.client.post(
            '/{}/upload-store/images/'.format(self.api_prefix),
            {'file': file}, format='multipart')
        response = self.client.get(response.data['url'])
        self.assertEqual(200, response.status_code, 'Image retrieving failed')


@patch('munch.core.mail.utils.get_app_url', mock_get_app_url)
class PermissionsAPITestCase(TestCase, HttpBasicTestCaseMixin):
    def setUp(self):
        self.api_prefix = 'v1'
        self.admin = UserFactory(
            is_admin=True, is_active=True, is_superuser=True)
        self.admin.organization.can_attach_files = False
        self.admin.organization.can_external_optout = False
        self.admin.organization.save()

    def create_random_image(self, width=800, height=600):
        color = (
            random.randint(0, 255),
            random.randint(0, 255),
            random.randint(0, 255))
        image = PIL.Image.new('RGB', (width, height), color)
        temp = NamedTemporaryFile(suffix='.png')
        image.save(temp, 'png')
        temp.seek(0)
        return temp

    def test_api_creation_permission(self):
        file = self.create_random_image()

        response = self.client.post(
            '/{}/upload-store/images/'.format(self.api_prefix),
            {'file': file}, format='multipart')
        self.assertEqual(401, response.status_code)

    def test_api_creation_permission_with_secret(self):
        file = self.create_random_image()
        auth_headers = self.http_basic_headers(
            'api', self.admin.secret)

        response = self.client.post(
            '/{}/upload-store/images/'.format(self.api_prefix), {'file': file},
            format='multipart', **auth_headers)
        self.assertEqual(201, response.status_code)

    def test_api_retrieve_permission(self):
        file = self.create_random_image()
        self.client.login(
            identifier=self.admin.identifier, password='password')
        response = self.client.post(
            '/{}/upload-store/images/'.format(self.api_prefix),
            {'file': file}, format='multipart')
        self.client.logout()
        response = self.client.get(response.data['url'])
        self.assertEqual(
            200, response.status_code,
            'Image retrieving failed while logged out')



from django.core import validators
from django.db import models
from django.db.models.signals import pre_save

import django_fsm


class EmailListField(models.CharField):
    """ Multiple emails

    after https://djangosnippets.org/snippets/3047/
    """
    class EmailListValidator(validators.EmailValidator):
        def __call__(self, value):
            if isinstance(value, (str, bytes)):
                value = value.split(',')
            for email in value:
                return super(
                    EmailListField.EmailListValidator, self).__call__(email)

    class Presentation(list):
        def __str__(self):
            return ",".join(self)

    default_validators = [EmailListValidator()]

    def get_db_prep_value(self, value, *args, **kwargs):
        if not value:
            return
        return ','.join(str(s) for s in value)

    def value_to_string(self, obj):
        value = self._get_val_from_obj(obj)
        return self.get_db_prep_value(value)

    def to_python(self, value):
        if not value:
            return
        elif isinstance(value, self.Presentation):
            return value
        else:
            return self.Presentation(
                [address.strip() for address in value.split(',')])


class FSMAutoField(django_fsm.FSMField):
    """ Pass transitions on attr change rather than transition calls

    Design of FSM is to call transitions explicitly.

    Here we take another approach, if an attribute is changed and then saved,
    we search for the first matching transition and apply it befor saving.

    If no matching transition is found, then we raise a TransitionNotAllowed.
    """

    @staticmethod
    def _fsm_check_transitions(sender, instance, raw, **kwargs):
        fieldname = instance._fsm_state_field
        Model = sender._meta.model
        # Excludes new instance or fixture loading
        if (instance.pk is not None) and (not raw):
            new = getattr(instance, fieldname)
            old = getattr(Model.objects.get(pk=instance.pk), fieldname)
            if new != old:
                field = Model._meta.get_field(fieldname)
                transitions = field.get_all_transitions(Model)

                relevant_transitions = filter(
                    lambda t: (t.target == new) and (t.source in (old, '*')),
                    transitions)
                try:
                    transition = list(relevant_transitions)[0]
                except IndexError:
                    raise django_fsm.TransitionNotAllowed(
                        '{} -> {} transition is not allowed'.format(old, new))
                else:
                    transition.method(instance)

    def contribute_to_class(self, cls, name, virtual_only=False):
        super().contribute_to_class(cls, name, virtual_only=False)
        setattr(cls, '_fsm_state_field', name)
        pre_save.connect(
            self._fsm_check_transitions, sender=cls,
            dispatch_uid='_fsm_check_transitions')



import logging

from celery import task

from .models import TrackRecord

log = logging.getLogger(__name__)


@task
def create_track_record(**kwargs):
    try:
        TrackRecord.objects.create(**kwargs)
    except:
        log.error('Error while tracking an event', exc_info=True)



import urllib
from email.utils import parseaddr
from email.header import make_header
from email.header import decode_header

from django.db import models
from django.urls import reverse
from django.conf import settings
from django.contrib.postgres.fields import HStoreField
from django.utils.translation import ugettext_lazy as _
from slimta.envelope import Envelope

from munch.core.mail.utils import mk_base64_uuid
from munch.core.mail.models import RawMail
from munch.core.mail.models import AbstractMail
from munch.core.mail.models import BaseMailQuerySet
from munch.core.mail.models import AbstractMailStatus
from munch.core.utils.models import AbstractOwnedModel
from munch.core.utils.managers import OwnedModelQuerySet

from munch.apps.users.models import MunchUser
from munch.apps.campaigns.models import Category
from munch.apps.domains.models import SendingDomain

from .munchers import headers_filters


class MailBatch(AbstractOwnedModel):
    creation_date = models.DateTimeField(
        verbose_name=_('creation date'), auto_now_add=True)
    identifier = models.CharField(
        max_length=35, db_index=True, unique=True, editable=False,
        default=mk_base64_uuid, verbose_name=_('identifier'))
    name = models.CharField(_('name'), max_length=60)
    author = models.ForeignKey(MunchUser, verbose_name=_('author'))
    category = models.ForeignKey(
        Category, null=True, blank=True,
        related_name='mailbatches', verbose_name=_('category'))
    msg_links = HStoreField(
        verbose_name=_('message links'), null=True, blank=True)

    owner_path = 'author__organization'
    author_path = 'author'

    class Meta:
        verbose_name = _('mail batch')
        verbose_name_plural = _('mail batches')
        unique_together = [('author', 'name')]

    def __str__(self):
        return '{}'.format(self.identifier)

    def get_organization(self):
        return self.author.organization

    def get_clicks_defaults(self):
        if self.msg_links:
            return {k: 0 for k in list(self.msg_links.values())}
        return {}

    def get_clicks_count(self, click_qs, unique=False):
        clicked = click_qs.count_by_url(
            self.msg_links, include_any=unique, unique=unique)
        links = self.get_clicks_defaults()
        links.update(clicked)
        return links

    def mk_stats(self):
        from munch.apps.tracking.models import TrackRecord
        from munch.apps.tracking.models import READ_BROWSER
        from munch.apps.optouts.models import OptOut

        mail_qs = self.mails.all()

        open_identifiers = self.mails.filter(
            track_open=True).values_list('identifier', flat=True)
        clicks_identifiers = self.mails.filter(
            track_clicks=True).values_list('identifier', flat=True)

        click_qs = TrackRecord.objects.filter(
            identifier__in=clicks_identifiers, kind='click')
        open_qs = TrackRecord.objects.filter(
            identifier__in=open_identifiers, kind='read').order_by(
            'identifier', 'creation_date')
        browser_views = open_qs.filter(
            properties__source=READ_BROWSER).distinct().count()
        optout_qs = OptOut.objects.filter(
            identifier__in=self.mails.values_list('identifier', flat=True))

        open_time_serie = [int(p[
            'reaction_time']) for p in open_qs.values_list(
            'properties', flat=True) if p['reaction_time']]
        open_median_time = None
        if open_time_serie:
            open_median_time = sum(open_time_serie) / len(open_time_serie)

        return {
            'count': mail_qs.info_counts(),
            'last_status': mail_qs.last_status_counts(),
            'tracking': {
                'opened': mail_qs.opened().count(),
                'open_median_time': open_median_time,
                'clicked': self.get_clicks_count(click_qs, unique=True),
                'clicked_total': self.get_clicks_count(click_qs, unique=False),
                'viewed_in_browser': browser_views
            },
            'timing': {
                'delivery_total': mail_qs.duration().seconds,
                'delivery_median': mail_qs.median('delivery_duration').seconds
            },
            'optout': optout_qs.count_by_origin(with_total=True)
        }


def get_mail_identifier():
    return mk_base64_uuid('t-')


class MailQuerySet(BaseMailQuerySet, OwnedModelQuerySet):
    pass


class Mail(AbstractOwnedModel, AbstractMail):
    identifier = models.CharField(
        max_length=35, db_index=True, unique=True,
        default=get_mail_identifier, verbose_name=_('identifier'))
    headers = HStoreField(verbose_name=_('headers'))
    message = models.ForeignKey(
        RawMail, on_delete=models.SET_NULL, verbose_name=_('message'),
        null=True, related_name='transactional_mail')
    sender = models.EmailField(verbose_name=_('sender'))
    author = models.ForeignKey(MunchUser, verbose_name=_('author'))
    batch = models.ForeignKey(
        MailBatch, null=True, blank=True,
        verbose_name=_('batch'), related_name='mails')
    track_open = models.BooleanField(
        _('track opens'), default=False,
        help_text=(_("Detect who opens message by adding hidden image.")))
    track_clicks = models.BooleanField(
        _('track clicks'), default=False,
        help_text=_(
            "Detect which links were clicked by replacing "
            "original links by redirections"))
    msg_links = HStoreField(
        verbose_name=_('message links'), null=True, blank=True)
    curstatus = models.CharField(
        max_length=20,
        default=AbstractMailStatus.UNKNOWN,
        choices=AbstractMailStatus._meta.get_field('status').choices,
        verbose_name=_('current status'))

    objects = BaseMailQuerySet.as_manager()

    owner_path = 'author__organization'
    author_path = 'author'

    @classmethod
    def unsubscribe_url(cls, identifier, author, sending_domain):
        from munch.core.mail.utils import get_app_url

        relative_url = reverse(
            'unsubscribe', kwargs={'identifier': identifier})
        organization_app_url = get_app_url(
            organization=author.organization, domain=sending_domain)
        return urllib.parse.urljoin(organization_app_url, relative_url)

    def get_sending_domain(self):
        all_domains = SendingDomain.objects
        org_id = self.get_organization()
        qs = all_domains.filter(organization=org_id) | all_domains.filter(
            alt_organizations__in=[org_id])
        organization_domains = qs.distinct()
        return organization_domains.get_from_email_addr(
            parseaddr(self.headers.get('From'))[1], must_raise=False)

    def get_app_url(self):
        from munch.core.mail.utils import get_app_url

        return get_app_url(
            domain=self.get_sending_domain(),
            organization=self.get_organization())

    def get_organization(self):
        return self.author.organization

    def get_category(self):
        if self.batch and self.batch.category:
            return self.batch.category

    @classmethod
    def get_envelope(cls, identifier):
        from .utils import get_envelope_from_identifier
        return get_envelope_from_identifier(identifier)

    def as_envelope(self, must_raise=True):
        envelope = Envelope()

        headers_filters.process(
            self.headers, self,
            settings.TRANSACTIONAL['HEADERS_FILTERS_PARAMS'])

        headers = ""
        for key, value in self.headers.items():
            if key in ('Subject', ):
                header = make_header(
                    decode_header(value), header_name=key, maxlinelen=78)
                value = header.encode(linesep='\r\n')
            headers += "{}: {}\n".format(key, value)

        message = ""
        if self.message:
            message = self.message.content or ""

        if self.message is None and must_raise:
            raise Exception(
                "Can't build this envelope because "
                "there is no RawMail attached to it.")
        envelope.parse(headers.encode('utf-8') + message.encode('utf-8'))
        envelope.sender = self.sender
        envelope.recipients.append(self.recipient)
        return envelope

    def get_original_links(self, links):
        original_links = {}
        for link in links:
            original_links[link] = self.msg_links.get(link)
        return original_links

    def last_status(self):
        try:
            # Try to use cached version, see api/views.py
            return self.last_status_cached[0]
        except IndexError:
            return None
        except AttributeError:
            try:
                return self.statuses.latest()
            except MailStatus.DoesNotExist:
                return None


class MailStatus(AbstractOwnedModel, AbstractMailStatus):
    mail = models.ForeignKey(
        Mail, related_name='statuses', verbose_name=_('mail'))

    owner_path = 'mail__author__organization'
    author_path = 'mail__author'

    class Meta:
        verbose_name = _('mail status')
        verbose_name_plural = _('mail statuses')
        get_latest_by = 'creation_date'

    def save(self, *args, **kwargs):
        super().save(*args, **kwargs)
        must_save_mail = False
        if self.status in self.FINAL_STATES:
            self.mail.message = None
            must_save_mail = True
        if self.status == AbstractMailStatus.DROPPED:
            self.mail.had_delay = True
            must_save_mail = True
        if must_save_mail:
            self.mail.save()

    def __str__(self):
        return _('Message {} is {}').format(self.mail.identifier, self.status)



from munch.core.utils.permissions import MunchResourcePermission


class ContactListMergePermission(MunchResourcePermission):
    """ Same permissions as writing on the ContactList """
    app_name = 'contacts'
    model_name = 'contactlist'



from django.test import TestCase
from django.conf import settings
from rest_framework.test import APIClient
from faker import Factory as FakerFactory

from munch.apps.users.tests.factories import UserFactory
from munch.apps.users.tests.factories import OrganizationFactory

from ...models import SendingDomain

faker = FakerFactory.create()


class SendingDomainAPITestCase(TestCase):
    def setUp(self):
        self.api_prefix = 'v1'
        self.organization = OrganizationFactory()
        self.admin = UserFactory(
            groups=['administrators'], organization=self.organization)
        self.manager = UserFactory(
            groups=['managers'], organization=self.organization)
        self.user = UserFactory(
            groups=['users'], organization=self.organization)

        self.client = APIClient()
        self.client.login(
            identifier=self.admin.identifier, password='password')

    def test_non_admin_cannot_create_sending_domain(self):
        data = {'name': 'example.com'}

        self.client.login(
            identifier=self.manager.identifier, password='password')
        response = self.client.post(
            '/{}/domains/'.format(self.api_prefix), data, format='json')
        self.assertEqual(403, response.status_code)

        self.client.login(
            identifier=self.user.identifier, password='password')
        response = self.client.post(
            '/{}/domains/'.format(self.api_prefix), data, format='json')
        self.assertEqual(403, response.status_code)

        self.client.login(
            identifier=self.admin.identifier, password='password')
        response = self.client.post(
            '/{}/domains/'.format(self.api_prefix), data, format='json')
        self.assertEqual(201, response.status_code)

    def test_non_admin_add_alt_organizations(self):
        organization = OrganizationFactory(parent=self.admin.organization)

        parent_url = self.client.get(
            '/{}/me/'.format(self.api_prefix)).data.get('organization')
        children = self.client.get(
            self.client.get(parent_url).data.get('_links')['children']['href'])


        data = {
            'name': 'example.com',
            'alt_organizations': [children.data[0]['url']]}

        self.client.login(
            identifier=self.manager.identifier, password='password')

        response = self.client.post(
            '/{}/domains/'.format(self.api_prefix), data, format='json')
        self.assertEqual(403, response.status_code)

        self.client.login(
            identifier=self.user.identifier, password='password')

        response = self.client.post(
            '/{}/domains/'.format(self.api_prefix), data, format='json')
        self.assertEqual(403, response.status_code)


    def test_admin_add_alt_organizations(self):
        organization = OrganizationFactory(parent=self.admin.organization)

        parent_url = self.client.get(
            '/{}/me/'.format(self.api_prefix)).data.get('organization')
        children = self.client.get(
            self.client.get(parent_url).data.get('_links')['children']['href'])


        data = {
            'name': 'example.com',
            'alt_organizations': [children.data[0]['url']]}

        self.client.login(
            identifier=self.admin.identifier, password='password')

        response = self.client.post(
            '/{}/domains/'.format(self.api_prefix), data, format='json')
        self.assertEqual(201, response.status_code)
        self.assertEqual(
            data['alt_organizations'], response.json()['alt_organizations'])

    def test_non_admin_list_domains(self):
        self.client.login(identifier=self.user.identifier, password='password')
        response = self.client.get('/{}/domains/'.format(self.api_prefix))
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.data.get('results')), 0)

        organization = OrganizationFactory(parent=self.admin.organization)

        parent_url = self.client.get(
            '/{}/me/'.format(self.api_prefix)).data.get('organization')
        children = self.client.get(
            self.client.get(parent_url).data.get('_links')['children']['href'])


        data = {
            'name': 'example.com',
            'alt_organizations': [children.data[0]['url']]}

        self.client.login(
            identifier=self.admin.identifier, password='password')

        self.client.post(
            '/{}/domains/'.format(self.api_prefix), data, format='json')

        self.client.login(identifier=self.user.identifier, password='password')
        response = self.client.get('/{}/domains/'.format(self.api_prefix))
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.data.get('results')), 1)
        self.assertEqual(
            response.data.get('results')[0]['name'], 'example.com')

    def test_admin_add_non_owned_organization(self):
        user = UserFactory(groups=['administrators'])

        self.client.login(identifier=user.identifier, password='password')
        parent_url = self.client.get(
            '/{}/me/'.format(self.api_prefix)).data.get('organization')


        data = {
            'name': 'another-example.com', 'alt_organizations': [parent_url]}

        self.client.login(
            identifier=self.admin.identifier, password='password')

        response = self.client.post(
            '/{}/domains/'.format(self.api_prefix), data, format='json')
        self.assertEqual(400, response.status_code)
        self.assertEqual(len(data['alt_organizations']), 1)
        self.assertEqual(SendingDomain.objects.count(), 0)



from email.utils import parseaddr

from slimta.smtp.reply import Reply
from slimta.queue import QueueError
from slimta.policy import QueuePolicy
from django.core.validators import validate_email
from django.core.exceptions import ValidationError

from munch.apps.domains.models import SendingDomain


class Check(QueuePolicy):
    def apply(self, envelope):
        sender_email = parseaddr(envelope.headers.get('From'))[1]

        try:
            validate_email(sender_email)
        except ValidationError:
            sender_email = None

        if not sender_email:
            error = QueueError()
            error.reply = Reply(
                code='554',
                message=(
                    '5.7.1 <{}>: Relay access denied: '
                    'Sending domain invalid').format(envelope.recipients[0]))
            raise error

        organization_domain = SendingDomain.objects.filter(
            organization=envelope.organization)
        domain = organization_domain.get_from_email_addr(
            sender_email, must_raise=False)

        if not domain:
            error = QueueError()
            error.reply = Reply(
                code='554',
                message=(
                    '5.7.1 <{}>: Relay access denied: Sending '
                    'domain for <{}> is not properly configured').format(
                        envelope.recipients[0], sender_email))
            raise error

        envelope.sending_domain = domain



import logging

from django.conf import settings
from django.utils.module_loading import import_string
from django.core.mail.backends.base import BaseEmailBackend

from munch.core.mail.utils import extract_domain
from munch.core.mail.utils import mk_base64_uuid
from munch.core.mail.models import AbstractMailStatus

log = logging.getLogger(__name__)


class DummyBackend(BaseEmailBackend):
    """ This is a dummy backend that doesn't send any mail """
    def __init__(
            self,
            mailstatus_class_path=None,
            record_status_task_path=None,
            *args, **kwargs):

        self.mailstatus_class_path = mailstatus_class_path
        self.record_status_task_path = record_status_task_path

        if self.mailstatus_class_path:
            self.mailstatus_class = import_string(self.mailstatus_class_path)
        if self.record_status_task_path:
            self.record_status_task = import_string(
                self.record_status_task_path)

    def _record_sending(self, identifier, recipient):
        if self.mailstatus_class_path:
            mailstatus = self.mailstatus_class(
                status=AbstractMailStatus.SENDING,
                destination_domain=extract_domain(recipient))
            self.record_status_task(
                mailstatus, identifier,
                'dummy-localhost')

    def send_envelope(self, envelope, attempts=0):
        self._record_sending(envelope.headers.get(
            settings.X_MESSAGE_ID_HEADER), envelope.recipients[0])

    def send_message(self, identifier, recipient, headers, attempts=0):
        self._record_sending(identifier, recipient)

    # To be used by Django as a standard email backend
    def send_messages(self, email_messages):
        if not email_messages:
            return
        for message in email_messages:
            self._record_sending(mk_base64_uuid(), message.to[0])
        return len(email_messages)

    def send_simple_message(self, message):
        self._record_sending(mk_base64_uuid(), message.to[0])

    def send_simple_envelope(self, envelope):
        self._record_sending(mk_base64_uuid(), envelope.recipients[0])


def get_backend():
    return import_string(settings.MASS_EMAIL_BACKEND)


def pre_save_mailstatus_signal(*args, **kwargs):
    base = settings.MASS_EMAIL_BACKEND.split('.')[0]
    import_path = '{}.signals.pre_save_mailstatus'.format(base)
    try:
        signal = import_string(import_path)
        return signal(*args, **kwargs)
    except ImportError as err:
        log.debug('Trying to import "{}", but failed...{}'.format(
            import_path, str(err)))


def post_save_mailstatus_signal(*args, **kwargs):
    base = settings.MASS_EMAIL_BACKEND.split('.')[0]
    import_path = '{}.signals.post_save_mailstatus'.format(base)
    try:
        signal = import_string(import_path)
        return signal(*args, **kwargs)
    except ImportError as err:
        log.debug('Trying to import "{}", but failed...{}'.format(
            import_path, str(err)))


Backend = get_backend()



from django import forms
from django.contrib.auth.forms import SetPasswordForm
from django.contrib.auth.forms import UserCreationForm

from .models import MunchUser


class MunchUserCreationForm(UserCreationForm):
    class Meta:
        model = MunchUser
        fields = ['identifier', 'first_name', 'last_name', 'organization']


class InvitationForm(SetPasswordForm):
    first_name = forms.CharField()
    last_name = forms.CharField()

    def save(self, commit=True):
        first_name = self.cleaned_data["first_name"]
        last_name = self.cleaned_data["last_name"]
        password = self.cleaned_data["new_password1"]
        self.user.set_password(password)
        if commit:
            self.user.is_active = True
            self.user.first_name = first_name
            self.user.last_name = last_name
            self.user.save()
        return self.user



import logging
import datetime

from django.db import models
from django.db.models import F
from django.db.models import Max
from django.db.models import Min
from django.db.models import Manager
from django.db.models import ExpressionWrapper

from munch.core.mail.models import BaseMailQuerySet
from munch.core.mail.models import BaseMailStatusQuerySet
from munch.core.utils.managers import OwnedModelQuerySet

log = logging.getLogger('munch')


class MailStatusQuerySet(OwnedModelQuerySet, BaseMailStatusQuerySet):
    pass


class MailManager(Manager):
    def create(self, *args, **kwargs):
        from .models import MailStatus

        mail = super().create(*args, **kwargs)
        MailStatus.objects.create(
            mail=mail, status=MailStatus.UNKNOWN,
            creation_date=mail.creation_date,
            raw_msg='Mail passed to infrastructure')
        return mail

    def bulk_create(self, objs, update_status=True, *args, **kwargs):
        from .models import Mail
        from .models import MailStatus

        # same enhancement as create, but on already instanciated objs
        for i in objs:
            i.identifier = Mail._meta.get_field('identifier').default()

        created_mails = super().bulk_create(objs, *args, **kwargs)

        # Get all Mails without an initial_status. May be improved if
        # https://code.djangoproject.com/ticket/19527 gets merged.

        # FYI: on a 1000 mails insert, the bulk_create of statuses takes ~0.8s
        # out of 4.8s
        if update_status:
            statuses = [MailStatus(
                mail=i, creation_date=i.creation_date,
                status=MailStatus.UNKNOWN,
                raw_msg='Mail passed to infrastructure')
                for i in self.filter(statuses=None)]

            MailStatus.objects.bulk_create(statuses)

        # Manually update mails cached attributs base on
        # `date`, `message` and cached fields as None.
        # Why manually and not with signal ?
        #   => https://git.owk.cc/crunchmail/munch/issues/2012#note_2757
        #
        # /!\ Warning /!\ This part will only work with bulk_create
        # for emails that have same `Message` instance.
        message, created_date = None, None
        if created_mails:
            message = getattr(created_mails[0], 'message')
            created_date = getattr(created_mails[0], 'creation_date')
        if message and created_date:
            self.model.objects.filter(
                creation_date=created_date, message=message,
                first_status_date=None, latest_status_date=None).update(
                    had_delay=False,
                    first_status_date=F('creation_date'),
                    latest_status_date=F('creation_date'),
                    curstatus=MailStatus.UNKNOWN,
                    delivery_duration=datetime.timedelta())

        return created_mails


class MailQuerySet(OwnedModelQuerySet, BaseMailQuerySet):
    def _legit_exclusionlist(
            self, message, include_bounces=False, include_optouts=False):
        from .models import OptOut

        exclusion_list = []

        if not include_bounces:
            exclusion_list += OptOut.objects.filter(
                origin=OptOut.BY_BOUNCE).values_list('address', flat=True)

        if not include_optouts:
            # If we define a category, the optouts are valid in this category
            # and thats all.
            if message.category:
                exclusion_list += OptOut.objects.exclude(
                    origin=OptOut.BY_BOUNCE).filter(
                        category=message.category).values_list(
                            'address', flat=True)
            else:
                organization = message.author.organization
                exclusion_list += OptOut.objects.exclude(
                    origin=OptOut.BY_BOUNCE).filter(
                        author__organization=organization).values_list(
                            'address', flat=True)

        return exclusion_list

    def legit_for(self, *args, **kwargs):
        return self.exclude(
            recipient__in=self._legit_exclusionlist(*args, **kwargs))

    def not_legit_for(self, *args, **kwargs):
        return self.filter(
            recipient__in=self._legit_exclusionlist(*args, **kwargs))

    def with_bounds(self):
        return self.annotate(
            start=Min('statuses__creation_date'),
            end=Max('statuses__creation_date')).annotate(
                statuses_delta=ExpressionWrapper(
                    F('end') - F('start'),
                    output_field=models.DurationField()
                ))



# -*- coding: utf-8 -*-
# Generated by Django 1.9.7 on 2016-07-06 10:09
from __future__ import unicode_literals

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import munch.core.utils.models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='Category',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=100, verbose_name='name')),
                ('author', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL, verbose_name='author')),
            ],
            options={
                'verbose_name': 'category',
                'verbose_name_plural': 'categories',
                'abstract': False,
                'default_permissions': ('view', 'add', 'change', 'delete', 'view_mine', 'change_mine', 'delete_mine', 'view_organizations', 'change_organizations', 'delete_organizations'),
            },
            bases=(models.Model, munch.core.utils.models.OwnedModelMixin),
        ),
        migrations.CreateModel(
            name='RawMail',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('signature', models.CharField(max_length=32, verbose_name='signature')),
                ('content', models.TextField(verbose_name='content')),
            ],
        ),
    ]



import math
import email
import unittest
import logging

import clamd

from django.test import TestCase
from django.conf import settings

from . import SpamChecker, SpamCheckerError, SpamResult


logging.disable(logging.CRITICAL)


SAMPLE_HAM = """Return-Path: Foo@example.com
Delivery-Date: Mon May 13 04:46:13 2013
Received: from mandark.labs.netnoteinc.com ([213.105.180.140]) by
    dogma.slashnull.org (8.11.6/8.11.6) with ESMTP id g4D3kCe15097 for
    <jm@jmason.org>; Mon, 13 May 2013 04:46:12 +0100
Message-ID: <53E4FAB7.903@example.com>
Date: Fri, 13 May 2013 04:40:39 +0100
From: Foo <Foo@example.com>
User-Agent: Mozilla/5.0 (X11; Linux x86_64; rv:24.0) Gecko/20100101 Icedove/24.6.0
MIME-Version: 1.0
To: Bar <Bar@example.com>
Subject: Reunion prochaine
Content-Type: text/plain; charset=UTF-8; format=flowed
Content-Transfer-Encoding: 8bit

Salut,

A bientôt pour une super reunion

Bien a toi,

--
   Foo
"""
SAMPLE_SPAM = """
Return-Path: merchantsworld2001@juno.com
Delivery-Date: Mon May 13 04:46:13 2002
Received: from mandark.labs.netnoteinc.com ([213.105.180.140]) by
    dogma.slashnull.org (8.11.6/8.11.6) with ESMTP id g4D3kCe15097 for
    <jm@jmason.org>; Mon, 13 May 2002 04:46:12 +0100
Received: from 203.129.205.5.205.129.203.in-addr.arpa ([203.129.205.5]) by
    mandark.labs.netnoteinc.com (8.11.2/8.11.2) with SMTP id g4D3k2D12605 for
    <jm@netnoteinc.com>; Mon, 13 May 2002 04:46:04 +0100
Received: from html (unverified [207.95.174.49]) by
    203.129.205.5.205.129.203.in-addr.arpa (EMWAC SMTPRS 0.83) with SMTP id
    <B0000178595@203.129.205.5.205.129.203.in-addr.arpa>; Mon, 13 May 2002
    09:04:46 +0530
Message-Id: <B0000178595@203.129.205.5.205.129.203.in-addr.arpa>
From: lmrn@mailexcite.com
To: ranmoore@cybertime.net
Subject: Real Protection, Stun Guns!  Free Shipping! Time:2:01:35 PM
Date: Mon, 28 Jul 1980 14:01:35
MIME-Version: 1.0
X-Keywords:
Content-Type: text/html; charset="DEFAULT"

<html>
<body>
<center>
<h3>
<font color="blue">
<b>
The Need For Safety Is Real In 2002, You Might Only Get One Chance - Be Ready!
<p>
Free Shipping & Handling Within The (USA) If You Order Before May 25, 2002!
<p>
3 Day Super Sale, Now Until May 7, 2002!  Save Up To $30.00 On Some Items!

</b>
</font>
</h3>
</center>
<p>
IT'S GETTING TO BE SPRING AGAIN, PROTECT YOURSELF AS YOU WALK,<br>
JOG AND EXERCISE OUTSIDE.  ALSO PROTECT YOUR LOVED ONES AS<br>
THEY RETURN HOME FROM COLLEGE!<br>
<p>
*     LEGAL PROTECTION FOR COLLEGE STUDENTS!<br>
*     GREAT UP'COMING OUTDOOR PROTECTION GIFTS!<br>
*     THERE IS NOTHING WORTH MORE PROTECTING THAN LIFE!<br>
*     OUR STUN DEVICES & PEPPER PRODUCTS ARE LEGAL PROTECTION!
<p>
<b>
<font color="red">
JOIN THE WAR ON CRIME!
</b>
</font>
<p>

STUN GUNS AND BATONS
<p>
EFFECTIVE - SAFE - NONLETHAL
<p>
PROTECT YOUR LOVED ONES AND YOURSELF
<p>
No matter who you are, no matter what City or Town you live in,<br>
if you live in America, you will be touched by crime.
<p>
You hear about it on TV.  You read about it in the newspaper.<br>
It's no secret that crime is a major problem in the U.S. today.<br>
Criminals are finding it easier to commit crimes all the time.
<p>
Weapons are readily available.  Our cities' police forces have<br>
more work than they can handle.  Even if these criminal are<br>
caught, they won't be spending long in our nation's overcrowded<br>
jails.  And while lawmakers are well aware of the crime problem,<br>
they don't seem to have any effective answers.
<p>
Our Email Address:  <a
href="mailto:Merchants4all@aol.com">Merchants4all@aol.com</a>
<p>
INTERESTED:
<p>
You will be protecting yourself within 7 days!  Don't Wait,<br>
visit our web page below, and join The War On Crime!
<p>
*****************<br>
<a
href="http://www.geocities.com/realprotection_20022003/">http://www.geocities.com/realprotection_20022003/</a><br>
*****************
<p>
Well, there is an effective answer.  Take responsibility for<br>
your own security.  Our site has a variety of quality personal<br>
security products.  Visit our site, choose the personal security<br>
products that are right for you.  Use them, and join the war on
crime!
<p>
FREE PEPPER SPRAY WITH ANY STUN UNIT PURCHASE.<br>
(A Value of $15.95)
<p>
We Ship Orders Within 5 To 7 Days, To Every State In The U.S.A.<br>
by UPS, FEDEX, or U.S. POSTAL SERVICE.  Visa, MasterCard, American<br>
Express & Debt Card Gladly Accepted.
<p>
Ask yourself this question, if you don't help your loved ones,
who will?
<p>
INTERESTED:
<p>
*****************<br>
<a
href="http://www.geocities.com/realprotection_20022003/">http://www.geocities.com/realprotection_20022003/</a><br>
*****************
<p>
___The Stun Monster 625,000 Volts ($86.95)<br>
___The Z-Force Slim Style 300,000 Volts ($64.95)<br>
___The StunMaster 300,000 Volts Straight ($59.95)<br>
___The StunMaster 300,000 Volts Curb ($59.95)<br>
___The StunMaster 200,000 Volts Straight ($49.95)<br>
___The StunMaster 200,000 Volts Curb ($49.95)<br>
___The StunBaton 500,000 Volts ($89.95)<br>
___The StunBaton 300,000 Volts ($79.95)<br>
___Pen Knife (One $12.50, Two Or More $9.00)<br>
___Wildfire Pepper Spray  (One $15.95, Two Or More $11.75)
<p>
___Add $5.75 For Shipping & Handling Charge.
<p>

To Order by postal mail, please send to the below address.<br>
Make payable to Mega Safety Technology.
<p>
Mega Safety Technology<br>
3215 Merrimac Ave.<br>
Dayton, Ohio  45405<br>
Our Email Address:  <a
href="mailto:Merchants4all@aol.com">Merchants4all@aol.com</a>
<p>
Order by 24 Hour Fax!!!  775-257-6657.
<p>
*****<br>
<b><font color="red">Important Credit Card Information! Please Read Below!</b></font>
 <br><br>
*     Credit Card Address, City, State and Zip Code, must match
      billing address to be processed.
<br><br>

CHECK____  MONEYORDER____  VISA____ MASTERCARD____ AmericanExpress___
Debt Card___
<br><br>
Name_______________________________________________________<br>
(As it appears on Check or Credit Card)
<br><br>
Address____________________________________________________<br>
(As it appears on Check or Credit Card)
<br><br>
___________________________________________________<br>
City,State,Zip(As it appears on Check or Credit Card)
<br><br>
___________________________________________________<br>
Country
<br><br>
___________________________________________________<br>
(Credit Card Number)
<br><br>
Expiration Month_____  Year_____
<br><br>
___________________________________________________<br>
Authorized Signature
<br><br>
<b>
*****IMPORTANT NOTE*****
</b>
<br><br>
If Shipping Address Is Different From The Billing Address Above,
Please Fill Out Information Below.
<br><br>
Shipping Name______________________________________________
<br><br>
Shipping Address___________________________________________
<br><br>
___________________________________________________________<br>
Shipping City,State,Zip
<br><br>
___________________________________________________________<br>
Country
<br><br>
___________________________________________________________<br>
Email Address & Phone Number(Please Write Neat)
</body>
</html>
"""


def get_spam_result_mock(msg, is_spam=False, score=0.0, checks=5):
    """
    @param msg : a string containing the message to check
    @param is_spam : a boolean that turn email as spam or not(default: False)
    @param score : a float that reflect spam checker score
    @param checks : a interger that reflect the number of checks done on email
    @return SpamResult
    """
    checks_separator = '\n\t*  '
    checks_examples = (
        '%0.1f NO_RELAYS Informational: message was not relayed via SMTP',
        '%0.1f WEIRD_PORT URI: Uses non-standard port number for HTTP',
        '%0.1f HTML_MESSAGE BODY: HTML included in message',
        '%0.1f MISSING_HEADERS Missing To: header',
        '%0.1f LOTS_OF_MONEY Huge... sums of money',
        '%0.1f MONEY_FORM Lots of money if you fill out a form',
        '%0.1f MISSING_DATE Missing Date: header',
        '%0.1f MISSING_MID Missing Message-Id: header',
        '%0.1f MISSING_SUBJECT Missing Subject: header',
        '%0.1f MISSING_FROM Missing From: header',
        '%0.1f NO_HEADERS_MESSAGE Message appears to be missing most RFC-822',
        '%0.1f FILL_THIS_FORM Fill in a form with personal information',
        '%0.1f NORMAL_HTTP_TO_IP URI: Uses a dotted-decimal IP address in URL',
        '%0.1f NO_RECEIVED Informational: message has no Received headers')
    spam_status_template = (
        '%s, score=%s required=5.0 tests=HTML_MESSAGE,NO_RECEIVED,'
        '\n\tNO_RELAYS,WEIRD_PORT autolearn=unavailable '
        'autolearn_force=no version=3.4.0')

    msg_as_email = email.message_from_bytes(msg.encode())

    if checks > len(checks_examples):
        raise Exception("SpamCheckerMock doesn't have enough checks examples")

    is_spam_str = "No"
    if is_spam:
        is_spam_str = "Yes"
        # Default `score` and `checks` for spam if not given
        if not score:
            score = 5.1
        if not checks:
            checks = 5
    msg_as_email.add_header(
        'X-Spam-Status', spam_status_template % (is_spam_str, score))

    # Generate X-Spam-Report based on `checks`
    report_header = ""
    # >>> math.modf(5.1)
    # >>> (0.09, 5.0)
    floating_score, rest_score = math.modf(score)
    for i in range(0, checks):
        check_score = rest_score / checks
        if i == 0:
            check_score += floating_score
        report_header += checks_separator
        report_header += checks_examples[i] % check_score
    msg_as_email.add_header('X-Spam-Report', report_header)
    return SpamResult.from_headers(msg_as_email)


def clamd_check_virus_mock(afile, is_virus=False):
    if afile.read() == clamd.EICAR or is_virus:
        return {'stream': ('FOUND', '')}
    return {'stream': ('NOT FOUND', '')}


class SpamCheckerDNSTest(TestCase):
    def test_invalid_host(self):
        with self.assertRaises(SpamCheckerError):
            sc = SpamChecker('inexistent.example.com')

    def test_norespond_host(self):
        with self.assertRaises(SpamCheckerError):
            sc = SpamChecker('example.com')

    def test_ok_host(self):
        sc = SpamChecker(host=settings.SPAMD_HOST, port=settings.SPAMD_PORT)
        self.assertIsInstance(sc.host, str)
        self.assertIsInstance(sc.port, int)


class SpamCheckerTest(TestCase):
    def setUp(self):
        self.sc = SpamChecker(host=settings.SPAMD_HOST, port=settings.SPAMD_PORT)

    def test_check_ham(self):
        with unittest.mock.patch(
                'munch.apps.spamcheck.SpamChecker.check',
                side_effect=lambda *args, **kwargs: get_spam_result_mock(
                    is_spam=False, score=2.4, checks=1, *args, **kwargs)):
            res = self.sc.check(SAMPLE_HAM)
        self.assertEqual(res.is_spam, False)
        self.assertEqual(res.score, 2.4)
        self.assertEqual(len(res.checks), 1)

    def test_check_spam(self):
        with unittest.mock.patch(
                'munch.apps.spamcheck.SpamChecker.check',
                side_effect=lambda *args, **kwargs: get_spam_result_mock(
                    is_spam=True, score=5.4, checks=10, *args, **kwargs)):
            res = self.sc.check(SAMPLE_SPAM)
        self.assertEqual(res.is_spam, True)
        self.assertEqual(res.score, 5.4)
        self.assertEqual(len(res.checks), 10)


class SpamResultTest(TestCase):
    def test_parsing(self):
        headers = {
            'X-Spam-Status': """Yes, score=9.8 required=5.0 tests=[DATE_IN_PAST_96_XX=2.6,
        FILL_THIS_FORM=0.001, FREEMAIL_ENVFROM_END_DIGIT=0.25,
        FREEMAIL_FORGED_FROMDOMAIN=0.001, FREEMAIL_FROM=0.001,
        HEADER_FROM_DIFFERENT_DOMAINS=0.001, HTML_MESSAGE=0.001, LOTS_OF_MONEY=0.001,
        MIME_HTML_ONLY=2.199, MONEY_FORM=2.301, RDNS_NONE=2.399] autolearn=disabled""",
            'X-Spam-Report':"""
        *  0.2 FREEMAIL_ENVFROM_END_DIGIT Envelope-from freemail username ends in
        *      digit (merchantsworld2001[at]juno.com)
        *  2.6 DATE_IN_PAST_96_XX Date: is 96 hours or more before Received: date
        *  0.0 FREEMAIL_FROM Sender email is commonly abused enduser mail provider
        *      (merchantsworld2001[at]juno.com)
        *  0.0 HEADER_FROM_DIFFERENT_DOMAINS From and EnvelopeFrom 2nd level mail
        *      domains are different""",
            'X-Spam-Flag': 'YES'
        }

        sr = SpamResult.from_headers(headers)
        self.assertEqual(sr.is_spam, True)
        self.assertEqual(sr.score, 9.8)
        self.assertEqual(len(sr.checks), 4)
        self.assertEqual(
            sr.checks[3]['description'],
            'From and EnvelopeFrom 2nd level mail domains are different')
        self.assertEqual(sr.checks[3]['score'], 0)
        self.assertEqual(sr.checks[3]['name'],'HEADER_FROM_DIFFERENT_DOMAINS')



# -*- coding: utf-8 -*-
from django.db import migrations
from django.core.management.sql import emit_post_migrate_signal

PERMISSIONS = {
    'organization': [
        ('add_organization', 'Can add organization'),
        ('change_mine_organization', 'Can change_mine organization'),
        ('change_organization', 'Can change organization'),
        ('change_organizations_organization', 'Can change_organizations organization'),
        ('delete_mine_organization', 'Can delete_mine organization'),
        ('delete_organization', 'Can delete organization'),
        ('delete_organizations_organization', 'Can delete_organizations organization'),
        ('view_mine_organization', 'Can view_mine organization'),
        ('view_organization', 'Can view organization'),
        ('view_organizations_organization', 'Can view_organizations organization'),
        ('add_child_organization', 'Can add child organizations'),
        ('delete_child_organization', 'Can delete child organizations'),
        ('invite_user_organization', 'Can invite user in organization'), ],
    'organizationsettings': [
        ('add_organizationsettings', 'Can add organization settings'),
        ('change_mine_organizationsettings', 'Can change_mine organization settings'),
        ('change_organizationsettings', 'Can change organization settings'),
        ('change_organizations_organizationsettings', 'Can change_organizations organization settings'),
        ('delete_mine_organizationsettings', 'Can delete_mine organization settings'),
        ('delete_organizationsettings', 'Can delete organization settings'),
        ('delete_organizations_organizationsettings', 'Can delete_organizations organization settings'),
        ('view_mine_organizationsettings', 'Can view_mine organization settings'),
        ('view_organizationsettings', 'Can view organization settings'),
        ('view_organizations_organizationsettings', 'Can view_organizations organization settings'), ],
    'munchuser': [
        ('add_munchuser', 'Can add user'),
        ('change_mine_munchuser', 'Can change_mine user'),
        ('change_munchuser', 'Can change user'),
        ('change_organizations_munchuser', 'Can change_organizations user'),
        ('delete_mine_munchuser', 'Can delete_mine user'),
        ('delete_munchuser', 'Can delete user'),
        ('delete_organizations_munchuser', 'Can delete_organizations user'),
        ('view_mine_munchuser', 'Can view_mine user'),
        ('view_munchuser', 'Can view user'),
        ('view_organizations_munchuser', 'Can view_organizations user'), ],
    'apiapplication': [
        ('add_apiapplication', 'Can add apiapplication'),
        ('change_apiapplication', 'Can change apiapplication'),
        ('change_mine_apiapplication', 'Can change_mine apiapplication'),
        ('change_organizations_apiapplication', 'Can change_organizations apiapplication'),
        ('delete_apiapplication', 'Can delete apiapplication'),
        ('delete_mine_apiapplication', 'Can delete_mine apiapplication'),
        ('delete_organizations_apiapplication', 'Can delete_organizations apiapplication'),
        ('view_apiapplication', 'Can view apiapplication'),
        ('view_mine_apiapplication', 'Can view_mine apiapplication'),
        ('view_organizations_apiapplication', 'Can view_organizations apiapplication'), ],
    'smtpapplication': [
        ('add_smtpapplication', 'Can add smtpapplication'),
        ('change_mine_smtpapplication', 'Can change_mine smtpapplication'),
        ('change_organizations_smtpapplication', 'Can change_organizations smtpapplication'),
        ('change_smtpapplication', 'Can change smtpapplication'),
        ('delete_mine_smtpapplication', 'Can delete_mine smtpapplication'),
        ('delete_organizations_smtpapplication', 'Can delete_organizations smtpapplication'),
        ('delete_smtpapplication', 'Can delete smtpapplication'),
        ('view_mine_smtpapplication', 'Can view_mine smtpapplication'),
        ('view_organizations_smtpapplication', 'Can view_organizations smtpapplication'),
        ('view_smtpapplication', 'Can view smtpapplication'), ],
}
GROUP_PERMISSIONS = {
    'administrators': {
        'organization': [
            'view_organizations_organization',
            'change_mine_organization',
            'change_organization',
            'add_child_organization',
            'delete_child_organization',
            'invite_user_organization'],
        'organizationsettings': [],
        'munchuser': [
            'add_munchuser',
            'change_organizations_munchuser',
            'view_organizations_munchuser', ],
        'apiapplication': [
            'add_apiapplication',
            'change_organizations_apiapplication',
            'delete_organizations_apiapplication',
            'view_organizations_apiapplication', ],
        'smtpapplication': [
            'add_smtpapplication',
            'change_organizations_smtpapplication',
            'delete_organizations_smtpapplication',
            'view_organizations_smtpapplication', ],
    },
    'managers': {
        'organization': [
            'view_organizations_organization', ],
        'organizationsettings': [],
        'munchuser': [
            'change_mine_munchuser',
            'view_mine_munchuser', ],
        'apiapplication': [
            'add_apiapplication',
            'change_organizations_apiapplication',
            'delete_organizations_apiapplication',
            'view_organizations_apiapplication', ],
        'smtpapplication': [
            'add_smtpapplication',
            'change_organizations_smtpapplication',
            'delete_organizations_smtpapplication',
            'view_organizations_smtpapplication', ],
    },
    'users': {
        'organization': [
            'view_organizations_organization', ],
        'organizationsettings': [],
        'munchuser': [
            'change_mine_munchuser',
            'view_mine_munchuser', ],
        'apiapplication': [],
        'smtpapplication': [],
    },
    'collaborators': {
        'organization': [
            'view_organizations_organization', ],
        'organizationsettings': [],
        'munchuser': [
            'change_mine_munchuser',
            'view_mine_munchuser', ],
        'apiapplication': [],
        'smtpapplication': [],
    },
}


def update_content_types(apps, schema_editor):
    db_alias = schema_editor.connection.alias
    emit_post_migrate_signal(False, 'default', db_alias)


def load_permissions(apps, schema_editor):
    Group = apps.get_model('auth', 'group')
    Permission = apps.get_model('auth', 'permission')
    ContentType = apps.get_model('contenttypes', 'contenttype')

    # Delete previous permissions
    for model in PERMISSIONS:
        content_type = ContentType.objects.get(
            app_label='users', model=model)
        Permission.objects.filter(content_type=content_type).delete()

    # Load permissions
    for model_name, permissions in PERMISSIONS.items():
        for permission_codename, permission_name in permissions:
            content_type = ContentType.objects.get(
                app_label='users', model=model_name)
            if not Permission.objects.filter(
                    codename=permission_codename,
                    content_type=content_type).exists():
                Permission.objects.create(
                    name=permission_name,
                    codename=permission_codename,
                    content_type=content_type)

    # Group permissions
    for group_name, models in GROUP_PERMISSIONS.items():
        group, _ = Group.objects.get_or_create(name=group_name)
        for model_name, permissions in models.items():
            content_type = ContentType.objects.get(
                app_label='users', model=model_name)
            for codename in permissions:
                permission, _ = Permission.objects.get_or_create(
                    codename=codename, content_type_id=content_type.id)
                group.permissions.add(permission)
                group.save()


class Migration(migrations.Migration):
    # TODO: Add your dependencies here:
    #   dependencies = [('users', '0001_initial')]
    dependencies = [('users', '0001_initial')]
    operations = [
        migrations.RunPython(update_content_types, reverse_code=None),
        migrations.RunPython(load_permissions, reverse_code=None)]
